import ocrmypdf

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = """
    Runs OCRmyPDF on a given a document (pdf, jpg, png or tiff).
    Creates associated svg, hocr and jpeg preview files.

    Example of usage:

    ./manage.py ocr ~/Desktop/file1.pdf --sidecar-dir ~/Desktop/file1 -l deu
    """

    def add_arguments(self, parser):
        parser.add_argument(
            'document',
            help="Input document. Can be PDF, jpg, png or tiff"
        )
        parser.add_argument(
            '-l',
            '--lang',
            default='deu',
            help="Language to OCR"
        )
        parser.add_argument(
            '--sidecar-dir',
            help="Folder where to write generated files"
        )
        parser.add_argument(
            '--sidecar-format',
            help="Format of generated output",
            choices=["html", "svg"],
            default="svg"
        )
        parser.add_argument(
            '--preview-width',
            help="Base width of preview image",
            type=int,
            default=400
        )
        parser.add_argument(
            '-k',
            '--keep-temporary-files',
            action='store_true',
            help="Keep temporary files"
        )

    def handle(self, *args, **options):
        document = options['document']
        sidecar_dir = options['sidecar_dir']
        sidecar_format = options['sidecar_format']
        lang = options['lang']
        keep = options['keep_temporary_files']
        output_document = f"{document.split('.')[0]}_output.pdf"
        preview_width = options['preview_width']

        ocrmypdf.ocr(
            document,
            output_document,
            lang=lang,
            plugins=["ocrmypdf_papermerge.plugin"],
            progress_bar=True,
            pdf_renderer='hocr',
            use_threads=True,
            keep_temporary_files=keep,
            sidecar_dir=sidecar_dir,
            sidecar_format=sidecar_format,
            preview_width=preview_width
        )



from datetime import timedelta

from rest_framework import status
from rest_framework.response import Response
from knox.models import AuthToken
from rest_framework_json_api.views import ModelViewSet
from rest_framework_json_api.renderers import JSONRenderer
from papermerge.core.serializers import (
    TokenSerializer,
    CreateTokenSerializer,
)
from papermerge.core.auth import CustomModelPermissions

from .mixins import RequireAuthMixin


class TokensViewSet(RequireAuthMixin, ModelViewSet):
    """
    Each user can have multiple authentication tokens. The reason to have
    multiple tokens per user is that he (or she) may consume REST API from
    multiple clients (or devices) using one single user account.
    User may then use a separate authentication token per each device or client.
    """
    queryset = AuthToken.objects.all()
    serializer_class = TokenSerializer
    renderer_classes = (JSONRenderer,)
    permission_classes = [CustomModelPermissions]

    http_method_names = ["get", "post", "delete", "head", "options"]

    def perform_create(self, serializer):
        return AuthToken.objects.create(
            user=self.request.user,
            expiry=timedelta(hours=serializer.data['expiry_hours'])
        )

    def create(self, request, *args, **kwargs):
        serializer = CreateTokenSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        instance, token = self.perform_create(serializer)
        authtoken_serializer = TokenSerializer(instance)
        headers = self.get_success_headers(authtoken_serializer.data)
        data = authtoken_serializer.data
        # return to user verbatim token only once - when token
        # was created
        data['token'] = token

        return Response(
            data,
            status=status.HTTP_201_CREATED,
            headers=headers
        )



import re

from django.core.exceptions import ValidationError


def safe_character_validator(value):

    unsafe_chars = re.compile(r'.*[<>\/\\;:&].*')
    message = 'Enter only safe characters.'

    if unsafe_chars.match(value):
        raise ValidationError(message)



from django.http import (
    JsonResponse,
    HttpResponse
)
from rest_framework.permissions import IsAuthenticated


class RequireAuthMixin:
    permission_classes = [IsAuthenticated]


class JSONResponseMixin:
    """
    A mixin that can be used to render a JSON response.
    """
    @property
    def asks_for_json(self):
        return any([
            'application/json' in self.request.headers.get('Content-Type', []),
            'application/json' in self.request.headers.get('Accept', [])
        ])

    def render_to_json_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        return JsonResponse(
            self.get_data(context),
            **response_kwargs
        )

    def get_data(self, context):
        """
        Returns an object that will be serialized as JSON by json.dumps().
        """
        # Note: This is *EXTREMELY* naive; in reality, you'll need
        # to do much more complex handling to ensure that arbitrary
        # objects -- such as Django model instances or querysets
        # -- can be serialized as JSON.
        return context


class HybridResponseMixin:
    """
    A mixin that can be used to render JSON or SVG responses.
    """

    @property
    def asks_for_json(self):
        return any([
            'application/json' in self.request.headers.get('Content-Type', []),
            'application/json' in self.request.headers.get('Accept', [])
        ])

    @property
    def asks_for_svg(self):
        return any([
            'image/svg' in self.request.headers.get('Content-Type', []),
            'image/svg' in self.request.headers.get('Accept', [])
        ])

    def render_to_json_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        return JsonResponse(
            self.get_data(context),
            **response_kwargs
        )

    def render_to_svg_response(self, svg_image, **response_kwargs):
        """
        Returns SVG response
        """
        return HttpResponse(
            svg_image,
            content_type='image/svg',
            **response_kwargs
        )

    def render_to_json_bad_request(self, error_message):
        return JsonResponse(
            {'error': error_message}, status=400
        )

    def get_data(self, context):
        """
        Returns an object that will be serialized as JSON by json.dumps().
        """
        # Note: This is *EXTREMELY* naive; in reality, you'll need
        # to do much more complex handling to ensure that arbitrary
        # objects -- such as Django model instances or querysets
        # -- can be serialized as JSON.
        return context


class GetClassSerializerMixin:
    def get_serializer_context(self):
        """
        Extra context provided to the serializer class.
        """
        return {
            'request': self.request,
            'format': self.format_kwarg,
            'view': self
        }

    def get_serializer_class(self):
        return self.class_serializer

    def get_serializer(self, *args, **kwargs):
        """
        Return the serializer instance that should be used for validating and
        deserializing input, and for serializing output.
        """
        serializer_class = self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)



import logging

from collections import OrderedDict

from django.http import (
    Http404,
    HttpResponse
)
from django.db.models.signals import post_save
from django.utils import encoding
from rest_framework.generics import (
    GenericAPIView,
    CreateAPIView,
    UpdateAPIView,
    DestroyAPIView
)
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer
from rest_framework_json_api.views import ModelViewSet
from rest_framework_json_api.parsers import JSONParser as JSONAPIParser
from rest_framework_json_api.renderers import JSONRenderer as JSONAPIRenderer
from rest_framework import status
from rest_framework.serializers import ListSerializer
from rest_framework.settings import api_settings
from rest_framework import relations

from rest_framework_json_api import utils


from drf_spectacular.utils import (
    extend_schema,
    OpenApiTypes
)

from papermerge.core.serializers.node import Data_NodeSerializer
from papermerge.core.serializers import (
    NodeSerializer,
    NodeMoveSerializer,
    NodesDownloadSerializer,
    NodeTagsSerializer,
    InboxCountSerializer
)

from papermerge.core.models import (
    BaseTreeNode,
    Document,
)
from papermerge.core.models.node import move_node

from papermerge.core.nodes_download import get_nodes_download

from .mixins import RequireAuthMixin, GetClassSerializerMixin

logger = logging.getLogger(__name__)


PER_PAGE = 30


class JSONAPIPolymorphicRenderer(JSONAPIRenderer):
    @classmethod
    def build_json_resource_obj(
        cls,
        fields,
        resource,
        resource_instance,
        resource_name,
        serializer,
        force_type_resolution=False,
    ):
        """
        When building JSON obj force correct node type! UGLY!
        """
        if resource_instance.is_folder:
            resource_name = "folders"
        else:
            resource_name = "documents"

        if resource_instance:
            resource_id = encoding.force_str(resource_instance.pk)
        else:
            resource_id = None

        resource_data = [
            ("type", resource_name),
            (
                "id",
                resource_id,
            ),
            ("attributes", cls.extract_attributes(fields, resource)),
        ]
        relationships = cls.extract_relationships(
            fields,
            resource,
            resource_instance
        )
        if relationships:
            relationships['parent']['data']['type'] = 'folders'  # UGLY!
            resource_data.append(("relationships", relationships))
        # Add 'self' link if field is present and valid
        if api_settings.URL_FIELD_NAME in resource and isinstance(
            fields[api_settings.URL_FIELD_NAME], relations.RelatedField
        ):
            resource_data.append(
                ("links", {"self": resource[api_settings.URL_FIELD_NAME]})
            )

        meta = cls.extract_meta(serializer, resource)
        if meta:
            resource_data.append(("meta", utils.format_field_names(meta)))

        return OrderedDict(resource_data)


class NodesViewSet(RequireAuthMixin, ModelViewSet):
    """
    Documents can be organized in folders. One folder can contain documents as
    well as other folders. A node is a convinient abstraction of two concepts -
    'folder' and 'document'. Each node has a type field with value either
    'folders' or 'documents' depending on what kind of node it is.
    """
    serializer_class = NodeSerializer
    parser_classes = [JSONAPIParser]
    renderer_classes = [JSONAPIPolymorphicRenderer]
    queryset = BaseTreeNode.objects.all()
    # object level permissions
    access_object_permissions = {
        'retrieve': "read",
        'update': "write",
        'delete': "delete"
    }

    @extend_schema(
        operation_id="node_retrieve",
        responses={
            200: ListSerializer(child=NodeSerializer())
        }
    )
    def retrieve(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def get_queryset(self, *args, **kwargs):
        """
        Retrieves node's children.

        Node is specified with pk = kwargs['pk'].
        If no pk is provided, retrieves all top level nodes of current user i.e.
        all nodes (of current user) with parent_id=None.
        """
        # This is workaround warning issued when runnnig
        # `./manage.py generateschema`
        # https://github.com/carltongibson/django-filter/issues/966
        if not self.request:
            return BaseTreeNode.objects.none()

        return BaseTreeNode.objects.filter(
            parent_id=self.kwargs.get('pk', None),
            user=self.request.user
        ).order_by('-created_at')

    @extend_schema(
        request=Data_NodeSerializer(),
        responses={
            201: Data_NodeSerializer(),
            400: {
                'type': 'object',
                'properties': {
                    'errors': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'detail': {
                                    'type': 'string'
                                },
                                'status': {
                                    'type': 'string'
                                },
                                'code': {
                                    'type': 'string'
                                },
                                'source': {
                                    'type': 'object',
                                    'properties': {
                                        'pointer': {
                                            'type': 'string'
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    )
    def create(self, request, *args, **kwargs):
        """
        Creates a node.

        A node can be either a Folder or a Document. In order to create
        a folder set required `type` attribute to `folders`. In order
        to create a document set `type` attribute to `documents`.

        Created document won't have any file associated i.e. this REST API
        creates just document model in database.
        """
        return super().create(request, *args, **kwargs)

    def perform_create(self, serializer):
        serializer.save(user_id=self.request.user.pk)


class NodesMoveView(RequireAuthMixin, GenericAPIView):
    parser_classes = [JSONParser]
    serializer_class = NodeMoveSerializer

    def post(self, request):
        serializer = NodeMoveSerializer(data=request.data)
        if serializer.is_valid():
            self._move_nodes(
                nodes=serializer.data['nodes'],
                target_parent_id=serializer.data['target_parent']['id']
            )
            return Response()
        else:
            return Response(
                serializer.errors,
                status=status.HTTP_400_BAD_REQUEST
            )

    def _move_nodes(self, nodes, target_parent_id):
        try:
            target_model = BaseTreeNode.objects.get(pk=target_parent_id)
        except BaseTreeNode.DoesNotExist as exc:
            logger.error(exc, exc_info=True)
            return

        for node in nodes:
            try:
                node_model = BaseTreeNode.objects.get(pk=node['id'])
            except BaseTreeNode.DoesNotExist as exc:
                logger.error(exc, exc_info=True)

            move_node(node_model, target_model)


class InboxCountView(RequireAuthMixin, APIView, GetClassSerializerMixin):
    parser_classes = [JSONParser]
    class_serializer = InboxCountSerializer

    def get(self, request):
        inbox_folder = request.user.inbox_folder
        return Response({
            'count': inbox_folder.get_children().count()
        })


class NodesDownloadView(RequireAuthMixin, GenericAPIView):
    """GET /nodes/download/"""
    parser_classes = [JSONParser]
    serializer_class = NodesDownloadSerializer

    @extend_schema(
        operation_id="nodes_download",
        responses={
            (200, 'application/pdf'): OpenApiTypes.BINARY,
            (200, 'application/zip'): OpenApiTypes.BINARY,
            (200, 'application/x-gtar'): OpenApiTypes.BINARY
        },
        parameters=[
            NodesDownloadSerializer
        ]
    )
    def get(self, request):
        serializer = NodesDownloadSerializer(data=request.query_params)
        if serializer.is_valid():
            try:
                nodes_download = get_nodes_download(**serializer.data)
            except Document.DoesNotExist as exc:
                raise Http404 from exc

            response = HttpResponse(
                nodes_download.get_content(),
                content_type=nodes_download.content_type
            )
            response['Content-Disposition'] = nodes_download.content_disposition

            return response
        else:
            return Response(
                serializer.errors,
                content_type='application/json',
                status=status.HTTP_400_BAD_REQUEST
            )

    def get_queryset(self):
        # This is workaround warning issued when runnnig
        # `./manage.py generateschema`
        # https://github.com/carltongibson/django-filter/issues/966
        if not self.request:
            return BaseTreeNode.objects.none()

        return super().get_queryset()


class NodeTagsView(
    RequireAuthMixin,
    CreateAPIView,
    UpdateAPIView,
    DestroyAPIView

):
    parser_classes = [JSONParser]
    renderer_classes = [JSONRenderer]
    serializer_class = NodeTagsSerializer
    pagination_class = None

    http_method_names = [
        'head',
        'post',
        'patch',
        'delete'
    ]

    def get_object(self):
        try:
            node = BaseTreeNode.objects.get(pk=self.kwargs['pk'])
        except BaseTreeNode.DoesNotExist as e:
            raise Http404("Node does not exist") from e

        return node

    @extend_schema(operation_id="node_assign_tags")
    def post(self, request, *args, **kwargs):
        """
        Assigns given list of tag names to the node.

        All tags not present in given list of tags names
        will be disassociated from the node; in other words upon
        successful completion of the request node will have ONLY
        tags from the list.
        Yet another way of thinking about http POST is as it **replaces
        existing node tags** with the one from input list.

        Example:

            Node N1 has 'invoice', 'important', 'unpaid' tags.

            After following request:

                POST /api/nodes/{N1}/tags/
                {tags: ['invoice', 'important', 'paid']}

            Node N1 will have 'invoice', 'important', 'paid' tags.
            Notice that previously associated 'unpaid' tag is not
            assigned to N1 anymore (because it was not in the provided list
            of tags).

        If you want to retain node tags not present in input tag list names
        then use PATCH/PUT http method of this endpoint.
        """
        ret = super().post(request, *args, **kwargs)
        # As the post_save signal is not sent automatically
        # send it manually. This signal is handled by
        # papermerge.search app in order to update the search
        # index
        post_save.send(
            sender=BaseTreeNode,
            instance=self.get_object(),
            created=False
        )
        return ret

    def perform_create(self, serializer):
        node = self.get_object().folder_or_document
        node.tags.set(
            serializer.data['tags'],
            tag_kwargs={"user": self.request.user}
        )

    @extend_schema(operation_id="node_append_tags")
    def patch(self, request, *args, **kwargs):
        """
        Appends given list of tag names to the node.

        Retains all previously associated node tags.
        Yet another way of thinking about http PATCH method is as it
        **appends** input tags to the currently associated tags.

        Example:

            Node N1 has 'invoice', 'important' tags.

            After following request:

                POST /api/nodes/{N1}/tags/
                {tags: ['paid']}

            Node N1 will have 'invoice', 'important', 'paid' tags.
            Notice that previously associated 'invoice' and 'important' tags
            are still assigned to N1.
        """
        return super().patch(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        node = self.get_object().folder_or_document

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        node.tags.add(
            *serializer.data['tags'],
            tag_kwargs={"user": self.request.user}
        )
        return Response(serializer.data)

    @extend_schema(operation_id="node_dissociate_tags")
    def delete(self, request, *args, **kwargs):
        """
        Dissociate given tags the node.

        Tags models are not deleted - just dissociated from the node.
        """
        return self.destroy(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        node = self.get_object().document_or_folder
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        node.tags.remove(*serializer.data['tags'])

        return Response(status=status.HTTP_204_NO_CONTENT)



import uuid
import logging

from django.utils.translation import gettext_lazy as _


logger = logging.getLogger(__name__)

OCR_STATUS_SUCCEEDED = 'succeeded'
OCR_STATUS_RECEIVED = 'received'
OCR_STATUS_STARTED = 'started'
OCR_STATUS_FAILED = 'failed'
OCR_STATUS_UNKNOWN = 'unknown'

OCR_STATUS_CHOICES = [
    ('unknown', _('Unknown')),
    ('received', _('Received')),
    ('started', _('Started')),
    ('succeeded', _('Succeeded')),
    ('failed', _('Failed')),
]


def uuid2raw_str(value: uuid.UUID) -> str:
    """Converts value into string as stored in database

    In database, UUID is stored as varchar(32) without '-' character.
    For example: UUID('1a606e93-b39c-439a-b8dd-8e981cb4d54b')
    will be converted to '1a606e93b39c439ab8dd8e981cb4d54b'.
    """
    return str(value).replace('-', '')



import pytz

from django.utils import timezone


class TimezoneMiddleware:
    """
    User preferences override l10n configuration from settings file.
    Per project (django project level) configurations are set in
    settings file (config.settings.*). Localization preferences
    i.e. for date & time format, timezone can be overridden by user.
    This middleware activates timezone considering user preferences.
    User preferences are set via:
        Admin -> Preferences -> Localization -> Timezone
    configuration.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):

        # user preferences make sense only
        # in case of authenticated user
        if request.user.is_anonymous:
            return self.get_response(request)

        preferences = request.user.preferences
        tzname = preferences['localization__timezone']

        if tzname:
            timezone.activate(pytz.timezone(tzname))
        else:
            timezone.deactivate()

        return self.get_response(request)



from django.urls import path, re_path
from django.conf.urls import include

from rest_framework import routers
from papermerge.core import views


router = routers.DefaultRouter()
router.include_root_view = False  # GET /api/ => results in 404

router.register(r"tokens", views.TokensViewSet, basename="token")
router.register(r"tags", views.TagsViewSet, basename="tag")
router.register("nodes", views.NodesViewSet, basename="node")
router.register(r"groups", views.GroupsViewSet)
router.register(r"users", views.UsersViewSet)
router.register(r"folders", views.FoldersViewSet, basename="folder")
router.register(
    r"documents", views.DocumentDetailsViewSet, basename="document"
)
router.register(
    r'preferences',
    views.CustomUserPreferencesViewSet,
    basename='preferences'
)

urlpatterns = [
    re_path(
        r'documents/(?P<document_id>[0-9a-f-]+)?/upload/(?P<file_name>[^/]+)',
        views.DocumentUploadView.as_view(),
        name='documents_upload'
    ),
    path(
        'documents/<uuid:pk>/ocr-text',
        views.DocumentOcrTextView.as_view(),
        name='document-ocr-text'
    ),
    path(
        'documents/merge/',
        views.DocumentsMergeView.as_view(),
        name='documents_merge'
    ),
    path(
        'document-versions/<uuid:pk>/download/',
        views.DocumentVersionsDownloadView.as_view(),
        name='download-document-version'
    ),
    path(
        'document-versions/<uuid:pk>/',
        views.DocumentVersionView.as_view(),
        name='document-version'
    ),
    path(
        'users/<uuid:pk>/change-password/',
        views.UserChangePassword.as_view(),
        name='users-change-password'
    ),
    path(
        'users/me/',
        views.CurrentUserView.as_view(),
        name='users-me'
    ),
    path(
        'nodes/move/',
        views.NodesMoveView.as_view(),
        name='nodes-move'
    ),
    path(
        'nodes/inboxcount/', views.InboxCountView.as_view(), name='inboxcount'
    ),
    path('nodes/download/', views.NodesDownloadView.as_view()),
    path(
        'nodes/<uuid:pk>/tags/',
        views.NodeTagsView.as_view(),
        name='node-tags'
    ),
    path(
        'pages/reorder/',
        views.PagesReorderView.as_view(),
        name='pages_reorder'
    ),
    path(
        'pages/rotate/',
        views.PagesRotateView.as_view(),
        name='pages_rotate'
    ),
    path(
        'pages/move-to-folder/',
        views.PagesMoveToFolderView.as_view(),
        name='pages_move_to_folder'
    ),
    path(
        'pages/move-to-document/',
        views.PagesMoveToDocumentView.as_view(),
        name='pages_move_to_document'
    ),
    path('pages/', views.PagesView.as_view(), name='pages'),  # only DELETE
    path(
        'pages/<uuid:pk>/',
        views.PageView.as_view(),
        name='pages_page'
    ),
    path('content-types/<int:pk>/', views.ContentTypeRetrieve.as_view()),
    path(
        'permissions/',
        views.PermissionsList.as_view(),
        name='permission-list'
    ),
    re_path(r'auth/login/', views.LoginView.as_view(), name='knox_login'),
    re_path(
        r'auth/logout/', views.LogoutView.as_view(), name='knox_logout'
    ),
    re_path(
        r'auth/logoutall/',
        views.LogoutAllView.as_view(),
        name='knox_logoutall'
    ),
    path(
        'ocr/',
        views.OCRView.as_view(),
        name='tasks-ocr'
    ),
    path(
        'version/',
        views.VersionView.as_view(),
        name='version'
    ),
    re_path(r"^", include(router.urls)),
]



# Generated by Django 3.2.13 on 2022-04-30 07:13

from django.conf import settings
import django.contrib.auth.models
import django.contrib.auth.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import papermerge.core.models.tags
import papermerge.core.validators
import taggit.managers
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=150, unique=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=150, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
                'abstract': False,
            },
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='BaseTreeNode',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=200, validators=[papermerge.core.validators.safe_character_validator], verbose_name='Title')),
                ('lang', models.CharField(default='deu', max_length=8, verbose_name='Language')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('lft', models.PositiveIntegerField(editable=False, null=True)),
                ('rght', models.PositiveIntegerField(editable=False, null=True)),
                ('tree_id', models.PositiveIntegerField(editable=False, null=True)),
                ('level', models.PositiveIntegerField(editable=False, null=True)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='core.basetreenode', verbose_name='parent')),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_core.basetreenode_set+', to='contenttypes.contenttype')),
            ],
            options={
                'verbose_name': 'Documents',
                'verbose_name_plural': 'Documents',
                'ordering': ('tree_id', 'lft'),
                'abstract': False,
                'base_manager_name': 'objects',
            },
        ),
        migrations.CreateModel(
            name='DocumentVersion',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('lang', models.CharField(default='deu', max_length=8, verbose_name='Language')),
                ('number', models.IntegerField(default=1, verbose_name='Version number')),
                ('file_name', models.CharField(blank=True, max_length=1024, null=True)),
                ('size', models.BigIntegerField(default=0, help_text='Size of file_orig attached. Size is in Bytes')),
                ('page_count', models.IntegerField(default=0)),
                ('short_description', models.CharField(blank=True, default='', max_length=128)),
                ('text', models.TextField(blank=True)),
            ],
            options={
                'verbose_name': 'Document version',
                'verbose_name_plural': 'Document versions',
                'ordering': ('number',),
            },
        ),
        migrations.CreateModel(
            name='KVStore',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(blank=True, max_length=200, null=True)),
                ('namespace', models.CharField(blank=True, max_length=1024, null=True)),
                ('namespaced_key', models.CharField(blank=True, max_length=200, null=True)),
                ('value', models.TextField(blank=True, null=True)),
                ('kv_type', models.CharField(choices=[('text', 'Text'), ('money', 'Currency'), ('numeric', 'Numeric'), ('date', 'Date')], default='text', max_length=16)),
                ('kv_format', models.CharField(blank=True, max_length=64, null=True)),
                ('kv_inherited', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='Document',
            fields=[
                ('basetreenode_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.basetreenode')),
                ('ocr', models.BooleanField(default=True)),
                ('ocr_status', models.CharField(choices=[('unknown', 'Unknown'), ('received', 'Received'), ('started', 'Started'), ('succeeded', 'Succeeded'), ('failed', 'Failed')], default='unknown', max_length=32)),
            ],
            options={
                'verbose_name': 'Document',
                'verbose_name_plural': 'Documents',
            },
            bases=('core.basetreenode',),
        ),
        migrations.CreateModel(
            name='Folder',
            fields=[
                ('basetreenode_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.basetreenode')),
            ],
            options={
                'verbose_name': 'Folder',
                'verbose_name_plural': 'Folders',
            },
            bases=('core.basetreenode',),
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('slug', models.SlugField(max_length=100, unique=True, verbose_name='slug')),
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('bg_color', models.CharField(default='#c41fff', max_length=7, verbose_name='Background Color')),
                ('fg_color', models.CharField(default='#FFFFFF', max_length=7, verbose_name='Foreground Color')),
                ('description', models.TextField(blank=True, max_length=1024, null=True, verbose_name='Description (optional)')),
                ('pinned', models.BooleanField(default=False, help_text='Pinned tag will be displayed under Documents menu. It serves as shortcut to quickly filter folders/documents associated with this tag')),
                ('name', models.CharField(max_length=100, validators=[papermerge.core.validators.safe_character_validator], verbose_name='name')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags',
                'ordering': ['name'],
                'unique_together': {('name', 'user')},
            },
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=64)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('permissions', models.ManyToManyField(blank=True, to='auth.Permission', verbose_name='permissions')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Page',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('number', models.IntegerField(default=1)),
                ('page_count', models.IntegerField(default=1)),
                ('text', models.TextField(default='')),
                ('norm_doc_title', models.CharField(default='', max_length=200)),
                ('norm_folder_title', models.CharField(default='', max_length=200)),
                ('norm_breadcrump', models.CharField(default='', max_length=1024)),
                ('norm_text', models.TextField(default='')),
                ('lang', models.CharField(default='deu', max_length=8)),
                ('image', models.CharField(default='', max_length=1024)),
                ('document_version', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pages', to='core.documentversion')),
            ],
            options={
                'ordering': ['number'],
            },
        ),
        migrations.CreateModel(
            name='KVStoreCompNode',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('node', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstorecomp', to='core.basetreenode')),
            ],
        ),
        migrations.CreateModel(
            name='ColoredTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.UUIDField(db_index=True, verbose_name='object ID')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='core_coloredtag_tagged_items', to='contenttypes.contenttype', verbose_name='content type')),
                ('tag', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='core_coloredtag_items', to='core.tag')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='basetreenode',
            name='tags',
            field=taggit.managers.TaggableManager(help_text='A comma-separated list of tags.', through='core.ColoredTag', to='core.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='basetreenode',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.CreateModel(
            name='Access',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('access_type', models.CharField(choices=[('allow', 'Allow'), ('deny', 'Deny')], default='allow', max_length=16)),
                ('access_inherited', models.BooleanField(default=False)),
                ('group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='auth.group', verbose_name='group')),
                ('node', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.basetreenode')),
                ('permissions', models.ManyToManyField(to='auth.Permission')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'permissions': [('read', 'Read Access'), ('write', 'Write Access'), ('delete', 'Delete Access'), ('change_perm', 'Change Permissions Access'), ('take_ownership', 'Take Ownership Access')],
            },
        ),
        migrations.AddField(
            model_name='user',
            name='role',
            field=models.ForeignKey(blank=True, help_text='Determines what operations this user is authorized to perform.', null=True, on_delete=django.db.models.deletion.CASCADE, to='core.role', verbose_name='role'),
        ),
        migrations.AddField(
            model_name='user',
            name='user_permissions',
            field=models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions'),
        ),
        migrations.CreateModel(
            name='KVStorePage',
            fields=[
                ('kvstore_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.kvstore')),
                ('page', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.page')),
            ],
            bases=('core.kvstore',),
        ),
        migrations.CreateModel(
            name='KVStoreNode',
            fields=[
                ('kvstore_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.kvstore')),
                ('node', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.basetreenode')),
            ],
            bases=('core.kvstore',),
        ),
        migrations.CreateModel(
            name='KVStoreCompItem',
            fields=[
                ('kvstore_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.kvstore')),
                ('comp_item', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='kvstore', to='core.kvstorecompnode')),
            ],
            bases=('core.kvstore',),
        ),
        migrations.AddField(
            model_name='documentversion',
            name='document',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='core.document', verbose_name='Document'),
        ),
        migrations.CreateModel(
            name='Automate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='Name')),
                ('match', models.CharField(blank=True, max_length=256, verbose_name='Match')),
                ('matching_algorithm', models.PositiveIntegerField(choices=[(1, 'Any'), (2, 'All'), (3, 'Literal'), (4, 'Regular Expression')], default=1, verbose_name='Matching Algorithm')),
                ('is_case_sensitive', models.BooleanField(default=True, verbose_name='Is case sensitive')),
                ('extract_page', models.BooleanField(default=False, verbose_name='Extract Page')),
                ('tags', papermerge.core.models.tags.UserTaggableManager(blank=True, help_text='A comma-separated list of tags.', through='core.ColoredTag', to='core.Tag', verbose_name='Tags')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('dst_folder', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='core.folder', verbose_name='Destination Folder')),
            ],
        ),
        migrations.AddField(
            model_name='user',
            name='home_folder',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='home_folder_of', to='core.folder'),
        ),
        migrations.AddField(
            model_name='user',
            name='inbox_folder',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='inbox_folder_of', to='core.folder'),
        ),
    ]



from django.contrib.auth.models import Permission

from rest_framework import generics
from rest_framework_json_api.renderers import JSONRenderer as JSONAPIRenderer

from papermerge.core.serializers import PermissionSerializer
from papermerge.core.auth import CustomModelPermissions
from .mixins import RequireAuthMixin


class PermissionsList(RequireAuthMixin, generics.ListAPIView):
    serializer_class = PermissionSerializer
    renderer_classes = (JSONAPIRenderer, )
    permission_classes = [CustomModelPermissions]
    pagination_class = None

    def get(self, request, *args, **kwargs):
        """
        Retrieves (almost) ALL permissions available in the system.

        It is useful to be able to retrieve (almost) ALL permission resource
        objects available - for example when creating a new Role you need to
        know what permissions are available and then select couple of available
        permissions and associate them to the newly create Role.

        Some permissions available will never be returned by this endpoint
        (hence "almost all").
        For example permissions for internally used 'BaseTreeNode' model.
        User is not aware what is 'BaseTreeNode' and getting permission
        for that will only confuse him/her. Permissions used to grant access
        per node/object are not returned either.

        In order to access this endpoint
        you need **view_permission** permission.

        Don't confuse this endpoint with "user permissions", which is intended,
        as name suggests, to return only user specific permissions.
        In order to retrieve current user permissions use
        /api/users/me/ endpoint.
        """
        return self.list(request, *args, **kwargs)

    def get_queryset(self):
        """
        Returns a queryset of permissions which makes sense
        to display (e.g. when creating a new Role).
        For example, it does not make sense to
        show permissions for BaseTreeNode model - user is not
        even aware of what is it (and should not be).
        """
        qs = Permission.objects.filter(
            content_type__model__in=[
                "user", "role", "group", "permission", "access",
                "authtoken", "session", "document",
                "folder", "userpreferencemodel", "automate"
            ]
        ).exclude(
            # exclude per node permissions,
            # i.e. permissions set on folder/document
            # instance (defined and
            # used in ``papermerge.core.access`` module)
            codename__in=[
                "read", "write", "delete", "change_perm", "take_ownership"
            ]
        )

        return qs



import logging
from django.core.management.base import BaseCommand

from papermerge.core.automate import apply_automates

logger = logging.getLogger(__name__)


class Command(BaseCommand):

    help = """
    Triggers core.automate.apply_automates(document_id, page_num).
    Useful for debugging purposes.
"""

    def add_arguments(self, parser):
        parser.add_argument(
            'document_id',
            type=int,
            help="Document id passed as argument to apply_automates function"
        )
        parser.add_argument(
            'page_num',
            type=int,
            help="page_num passed as argument to apply_automates function."
        )

    def handle(self, *args, **options):
        document_id = options.get(
            'document_id',
            False
        )
        page_num = options.get(
            'page_num',
            False
        )

        apply_automates(document_id, page_num)



from django.contrib.auth.models import ContentType

from rest_framework_json_api import serializers


class ContentTypeSerializer(serializers.ModelSerializer):

    class Meta:
        model = ContentType
        resource_name = 'content-types'
        fields = ('id', 'model',)



from rest_framework import serializers
from dynamic_preferences.users.serializers import UserPreferenceSerializer
from drf_spectacular.utils import extend_schema_field
from drf_spectacular.openapi import OpenApiTypes


class CustomUserPreferenceSerializer(UserPreferenceSerializer):
    id = serializers.CharField(read_only=True)

    class Meta:
        resource_name = 'preferences'
        fields = [
            'id',
            'default',
            'value',
            'help_text',
        ]

    @extend_schema_field(OpenApiTypes.STR)
    def get_default(self, o):
        return o.preference.api_repr(o.preference.get("default"))

    @extend_schema_field(OpenApiTypes.STR)
    def get_identifier(self, o):
        return o.preference.identifier()

    @extend_schema_field(OpenApiTypes.STR)
    def get_help_text(self, o):
        return o.preference.get("help_text")

    @extend_schema_field(OpenApiTypes.STR)
    def get_additional_data(self, o):
        return o.preference.get_api_additional_data()

    @extend_schema_field(OpenApiTypes.STR)
    def get_field(self, o):
        return o.preference.get_api_field_data()



from knox.views import LogoutView as KnoxLogoutView
from knox.views import LogoutAllView as KnoxLogoutAllView


class LogoutView(KnoxLogoutView):
    """
    Logs the user out of current session.

    On a successful request, the token used to authenticate is deleted from the
    system and can no longer be used for authentication.

    Request body must be empty.
    """
    def get_serializer(self, *args, **kwargs):
        return None


class LogoutAllView(KnoxLogoutAllView):
    """
    Logs the user out of all sessions i.e. deletes all auth tokens for the user.

    On a successful request, the token used to authenticate, and all other
    tokens registered to the same user account, are deleted from the system
    and can no longer be used for authentication.

    Request body must be empty.
    """
    def get_serializer(self, *args, **kwargs):
        return None



import logging

from django.utils.translation import gettext as _

from .lib.path import PagePath, DocumentPath
from .lib.pagecount import get_pagecount
from .models import Document, Automate
from .storage import default_storage
from .signal_definitions import automates_matching


logger = logging.getLogger(__name__)


def apply_automates(document_id, page_num, **kwargs):

    version = kwargs.get('version', 0)
    logger.debug("apply_automates: Begin.")
    try:
        document = Document.objects.get(id=document_id)
    except Document.DoesNotExist:
        logger.error(f"Provided document_id={document_id}, does not exists")
        return

    # use text files from the original version of the document
    doc_path = DocumentPath.copy_from(
        document.path(version=version)
    )
    page_count = get_pagecount(
        default_storage.abspath(doc_path.url())
    )
    page_path = PagePath(
        document_path=doc_path,
        page_num=page_num,
        page_count=page_count,
    )
    user = document.user

    text_path = default_storage.abspath(page_path.txt_url())
    text = ""
    with open(text_path, "r") as f:
        text = f.read()

    automates = Automate.objects.filter(user=user)
    # are there automates for the user?
    if automates.count() == 0:
        logger.debug(
            f"No automates for user {user}. Quit."
        )
        return

    # check all automates for given user (the owner of the document)
    matched = []
    for automate in automates:
        if automate.is_a_match(text):
            logger.debug(f"Automate {automate} matched document={document}")

            automate.apply(
                document=document,
                page_num=page_num,
                text=text,
            )
            matched.append(automate)
        else:
            logger.debug(
                f"No match for automate={automate}"
                f" doc_id={document_id} page_num={page_num}"
            )

    message = ""

    message = _(
        "%(count)s of %(total)s Automate(s) matched. ") % {
        'count': len(matched),
        'total': automates.count()
    }

    if len(matched) > 0:
        message += _("List of matched Automates: %(matched_automates)s") % {
            'matched_automates': matched
        }

    automates_matching.send(
        sender="papermerge.core.automate",
        user_id=document.user.id,
        document_id=document_id,
        level=logging.INFO,
        message=message,
        page_num=page_num,
        text=text,
        **kwargs
    )



from rest_framework import serializers as rest_serializers


class OcrSerializer(rest_serializers.Serializer):
    doc_id = id = rest_serializers.CharField(max_length=32, required=True)
    lang = rest_serializers.CharField(required=True)



import logging
import os
import uuid

from django.db import models

from papermerge.core.lib.path import PagePath
from papermerge.core.storage import abs_path
from papermerge.core.utils import clock

from .utils import (
    OCR_STATUS_SUCCEEDED,
    OCR_STATUS_UNKNOWN
)

logger = logging.getLogger(__name__)


class Page(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)

    document_version = models.ForeignKey(
        to='DocumentVersion',
        on_delete=models.CASCADE,
        related_name='pages'
    )

    number = models.IntegerField(default=1)
    page_count = models.IntegerField(default=1)

    text = models.TextField(default='')

    # inherited/normalized title from parent document
    norm_doc_title = models.CharField(
        max_length=200,
        default=''
    )

    # inherited/normalized title of immediate parent folder
    norm_folder_title = models.CharField(
        max_length=200,
        default=''
    )

    # normalized space delimited path (by folder title) of parent folder
    norm_breadcrump = models.CharField(
        max_length=1024,
        default=''
    )

    # text from all pages of the document
    norm_text = models.TextField(default='')

    # hm, this one should be norm_lang as well
    lang = models.CharField(
        max_length=8,
        blank=False,
        null=False,
        default='deu'
    )

    image = models.CharField(
        max_length=1024,
        default=''
    )

    class Meta:
        # Guarantees that
        # doc.pages.all() will return pages ordered by number.
        # test by
        # test_page.TestPage.test_pages_all_returns_pages_ordered
        ordering = ['number']

    def __str__(self):
        return f"id={self.pk} number={self.number}"

    @property
    def is_archived(self):
        """
        Returns True of page is archived.

        Page is considered archived if it belongs to archived document version.
        In other words, page is considered archived it is part of non-last
        document version.
        """
        return self.document_version.is_archived

    @property
    def is_last(self):
        return self.number == self.page_count

    @property
    def is_first(self):
        return self.number == 1

    @property
    def page_path(self):

        return PagePath(
            document_path=self.document_version.document_path,
            page_num=self.number,
        )

    @property
    def has_text(self):
        return len(self.stripped_text) != 0

    @property
    def stripped_text(self):
        return self.text.strip()

    def update_text_field(self, stream):
        """Update text field from given IO stream.

        Returns text read from IO stream
        """
        logger.debug(
            'update_text_field:'
            f'len(page.stripped_text)=={len(self.stripped_text)}'
        )
        self.text = stream.read()
        self.save()

        return self.stripped_text

    @property
    def txt_url(self):
        result = PagePath(
            document_path=self.document_version.document_path,
            page_num=self.number
        )

        return result.txt_url

    @property
    def txt_exists(self):

        result = PagePath(
            document_path=self.document.document_path,
            page_num=self.number
        )

        return result.txt_exists()

    def norm(self):
        """shortcut normalization method"""
        self.normalize_doc_title()
        self.normalize_folder_title()
        self.normalize_breadcrump()
        self.normalize_text()
        self.normalize_lang()

    def normalize_doc_title(self):
        """
        Save containing document's title
        """
        self.norm_doc_title = self.document.title
        self.save()

    def normalize_folder_title(self):
        """
        Save direct parent folder (containing folder) title
        """
        if self.document.parent:
            self.norm_folder_title = self.document.parent.title
            self.save()

    def normalize_breadcrump(self):
        pass

    def normalize_text(self):
        pass

    def normalize_lang(self):
        pass

    def get_ocr_status(self):
        """
        Returns OCR status of the document.

        Document model knows only limited information about
        document OCR status. From point of view of the document
        OCR status can be one of following:

            * succeeded - when document.text field is non empty
            * unknown - when document.text is empty

        In case of "unknown" OCR status application will need to query
        different parts of the system to figure out more details
        about OCR status.
        """
        if len(self.text) > 0:
            return OCR_STATUS_SUCCEEDED

        return OCR_STATUS_UNKNOWN

    @clock
    def get_jpeg(self):
        jpeg_abs_path = abs_path(self.page_path.preview_url)

        if not os.path.exists(jpeg_abs_path):
            # generate preview only for this page
            self.document_version.generate_previews(
                page_number=self.number
            )

        if not os.path.exists(jpeg_abs_path):
            # means that self.generate_preview() failed
            # to extract page image from the document
            raise IOError

        with open(jpeg_abs_path, "rb") as f:
            data = f.read()

        return data

    @clock
    def get_svg(self):
        svg_abs_path = abs_path(
            self.page_path.svg_url
        )

        if not os.path.exists(svg_abs_path):
            raise IOError

        with open(svg_abs_path, "rb") as f:
            data = f.read()

        return data


def get_pages(
    nodes,
    include_pages_with_empty_text=True
):
    """
    :nodes: is a query set of BaseTreeNodes

    Returns a QuerySet of Page(s)
    """
    if nodes.count() == 0:
        return Page.objects.none()

    pages = Page.objects.filter(
        document__in=nodes
    )

    if not include_pages_with_empty_text:
        ret_pages = pages.exclude(
            text__isnull=True
        ).exclude(
            text__exact=''
        )
    else:
        ret_pages = pages

    return ret_pages



from django.apps import AppConfig


class CoreConfig(AppConfig):
    name = 'papermerge.core'
    label = 'core'

    def ready(self):
        from papermerge.core import signals  # noqa
        from papermerge.core import checks  # noqa
        from papermerge.core.openapi import schema  # noqa



from django.utils.encoding import smart_str
from rest_framework import renderers


class PlainTextRenderer(renderers.BaseRenderer):
    media_type = 'text/plain'
    format = 'txt'

    def render(self, data, media_type=None, renderer_context=None):
        return smart_str(data, encoding=self.charset)


class ImageJpegRenderer(renderers.BaseRenderer):
    media_type = 'image/jpeg'
    format = 'jpeg'
    render_style = 'binary'

    def render(self, data, media_type=None, renderer_context=None):
        return data


class ImageSVGRenderer(renderers.BaseRenderer):
    media_type = 'image/svg+xml'
    format = 'svg'

    def render(self, data, media_type=None, renderer_context=None):
        return data


class PDFRenderer(renderers.BaseRenderer):
    media_type = 'application/pdf'
    format = 'pdf'

    def render(self, data, media_type=None, renderer_context=None):
        return data



# Generated by Django 4.0.5 on 2022-07-04 10:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0002_auto_20220619_0847'),
    ]

    operations = [
        migrations.AlterField(
            model_name='basetreenode',
            name='polymorphic_ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_%(app_label)s.%(class)s_set+', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='coloredtag',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='%(app_label)s_%(class)s_tagged_items', to='contenttypes.contenttype', verbose_name='content type'),
        ),
        migrations.AlterField(
            model_name='coloredtag',
            name='tag',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='%(app_label)s_%(class)s_items', to='core.tag'),
        ),
        migrations.AlterField(
            model_name='tag',
            name='slug',
            field=models.SlugField(allow_unicode=True, max_length=100, unique=True, verbose_name='slug'),
        ),
    ]



from django.conf import settings as django_settings


class AppSettings():
    """
    papermerge.core settings and their defaults
    """

    def __init__(self, prefix):
        self.prefix = prefix

    def _settings(self, name, default_value):
        full_name = f"{self.prefix}_{name}"
        value = getattr(
            django_settings,
            full_name,
            default_value,
        )
        return value

    @property
    def DEFAULT_FILE_STORAGE(self):  # noqa
        return self._settings(
            "DEFAULT_FILE_STORAGE",
            "papermerge.core.lib.storage.FileSystemStorage"
        )

    @property
    def FILE_STORAGE_KWARGS(self):  # noqa
        return self._settings(
            "FILE_STORAGE_KWARGS",
            {}
        )

    @property
    def BINARY_FILE(self):  # noqa
        return self._settings(
            "BINARY_FILE",
            "/usr/bin/file"
        )

    @property
    def BINARY_CONVERT(self):  # noqa
        return self._settings(
            "BINARY_CONVERT",
            "/usr/bin/convert"
        )

    @property
    def BINARY_IDENTIFY(self):  # noqa
        return self._settings(
            "BINARY_IDENTIFY",
            "/usr/bin/identify"
        )

    @property
    def BINARY_OCR(self):  # noqa
        return self._settings(
            "BINARY_OCR",
            "/usr/bin/tesseract"
        )

    @property
    def CONFIG_ENV_NAME(self):  # noqa
        """
        Name of environment variable pointing to
        papermerge.conf.py file.

        Useful in case papermerge.conf.py file is neither
        in root folder nor in /etc/papermerge.conf.py
        """
        return self._settings(
            "DEFAULT_CONFIG_ENV_NAME",
            "PAPERMERGE_CONFIG"
        )


settings = AppSettings(prefix="PAPERMERGE")



from rest_framework_json_api import serializers
from rest_framework.fields import empty
from rest_framework_json_api.utils import format_field_name
from drf_spectacular.extensions import OpenApiSerializerExtension
from drf_spectacular.extensions import OpenApiAuthenticationExtension


class KnoxTokenScheme(OpenApiAuthenticationExtension):
    target_class = 'knox.auth.TokenAuthentication'
    name = 'Token Authentication'

    def get_security_definition(self, auto_schema):
        return {
            'type': 'apiKey',
            'in': 'header',
            'name': 'Authorization',
            'description':
                'Token-based authentication with required prefix Token'
        }


class GroupJsonAPISerializer(OpenApiSerializerExtension):
    target_class = 'papermerge.core.serializers.GroupSerializer'
    type = {
        "type": "string",
        "enum": ["groups"]
    }

    def map_serializer(self, auto_schema, direction):
        required = []
        attributes = {}
        relationships = {}
        for field in self.target.fields.values():
            if isinstance(field, serializers.HyperlinkedIdentityField):
                # the 'url' is not an attribute but rather a self.link,
                # so don't map it here.
                continue
            if isinstance(field, serializers.HiddenField):
                continue
            if isinstance(field, serializers.RelatedField):
                relationships[format_field_name(field.field_name)] = {
                    "$ref": "#/components/schemas/reltoone"
                }
                continue
            if isinstance(field, serializers.ManyRelatedField):
                relationships[format_field_name(field.field_name)] = {
                    "$ref": "#/components/schemas/reltomany"
                }
                continue

            if field.required:
                required.append(field.field_name)

            schema = auto_schema._map_serializer_field(field, direction)
            if field.read_only:
                schema["readOnly"] = True
            if field.write_only:
                schema["writeOnly"] = True
            if field.allow_null:
                schema["nullable"] = True
            if field.default and field.default != empty:
                schema["default"] = field.default
            if field.help_text:
                # Ensure django gettext_lazy is rendered correctly
                schema["description"] = str(field.help_text)

            auto_schema._insert_field_validators(field, schema)

            attributes[format_field_name(field.field_name)] = schema

        result = {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": self.type,
                "id": {"type": "string", "format": "uuid"}
            },
        }
        if attributes:
            result["properties"]["attributes"] = {
                "type": "object",
                "properties": attributes,
            }
            if required:
                result["properties"]["attributes"]["required"] = required

        if relationships:
            result["properties"]["relationships"] = {
                "type": "object",
                "properties": relationships,
            }

        return result


class TagJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        "enum": ["tags"]
    }
    target_class = 'papermerge.core.serializers.TagSerializer'


class RolesJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        "enum": ["roles"]
    }
    target_class = 'papermerge.core.serializers.RoleSerializer'


class UserJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        "enum": ["users"]
    }
    target_class = 'papermerge.core.serializers.UserSerializer'


class DocumentDetailsJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        "enum": ["documents"]
    }
    target_class = 'papermerge.core.serializers.DocumentDetailsSerializer'


class FolderJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        "enum": ["folders"]
    }
    target_class = 'papermerge.core.serializers.FolderSerializer'


class NodeJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        # TODO: change "Document" to "documents" to be consistent with other
        # resource names
        "enum": ["Document", "documents", "folders"]

    }
    target_class = 'papermerge.core.serializers.NodeSerializer'


class PageJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        "enum": ["pages"]
    }
    target_class = 'papermerge.core.serializers.PageSerializer'


class TokenJsonAPISerializer(GroupJsonAPISerializer):
    type = {
        "type": "string",
        "enum": ["tokens"]
    }
    target_class = 'papermerge.core.serializers.TokenSerializer'


class CustomUserPreferenceSerializer(OpenApiSerializerExtension):

    target_class = 'papermerge.core.serializers.CustomUserPreferenceSerializer'

    def map_serializer(self, auto_schema, direction):
        response = {
            'type': 'object',
            'properties': {
                'id': {
                    'type': 'string',
                },
                'type': {
                    'type': 'string'
                },
                'attributes': {
                    'type': 'object',
                    'properties': {
                        'section': {
                            'type': 'string',
                            'readOnly': True
                        },
                        'identifier': {
                            'type': 'string',
                            'readOnly': True
                        },
                        'default': {
                            'type': 'string',
                            'readOnly': True
                        },
                        'help_text': {
                            'type': 'string',
                            'readOnly': True,
                            'nullable': True
                        },
                        'value': {
                            'type': 'string',
                            'readOnly': True
                        },
                        'name': {
                            'type': 'string',
                            'readOnly': True
                        }
                    }
                }
            }
        }

        return response



from rest_framework.renderers import JSONRenderer as rest_framework_JSONRenderer
from rest_framework.parsers import JSONParser as rest_framework_JSONParser
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response

from papermerge.core.serializers import (
    VersionSerializer
)
from papermerge.core.version import __version__ as THE_VERSION

from .mixins import RequireAuthMixin


class VersionView(RequireAuthMixin, GenericAPIView):
    serializer_class = VersionSerializer
    parser_classes = (rest_framework_JSONParser,)
    renderer_classes = (rest_framework_JSONRenderer,)

    def get(self, request):
        """
        Retrieves papermerge core module version
        """
        return Response(data={'version': THE_VERSION})



from django.contrib.auth import login

from rest_framework import permissions
from rest_framework.renderers import JSONRenderer
from rest_framework.parsers import JSONParser

from knox.views import LoginView as KnoxLoginView
from drf_spectacular.utils import extend_schema

from papermerge.core.serializers.login import (
    AuthTokenRequestSerializer,
    AuthTokenResponseSerializer
)
from .mixins import GetClassSerializerMixin


class LoginView(KnoxLoginView, GetClassSerializerMixin):
    """
    Authenticates user with given username and password.
    Response will contain token to be used as part of ``Authorization`` header
    in subsequent requests whenever authorization is required.
    """
    permission_classes = (permissions.AllowAny,)
    renderer_classes = (JSONRenderer,)
    parser_classes = (JSONParser,)
    class_serializer = AuthTokenRequestSerializer

    @extend_schema(
        request=AuthTokenRequestSerializer,
        responses={200: AuthTokenResponseSerializer},
    )
    def post(self, request, format=None):
        serializer = AuthTokenRequestSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data['user']
        login(request, user)
        return super(LoginView, self).post(request, format=None)



from configula import Configula

"""
Tesseract uses ISO-639-2/T for language names
 key is ISO-639-2/T as per:
 https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
"""

# We can't read configuration from django settings here, because API from this
# module i.e. ``get_ocr_langs`` and ``get_default_ocr_lang`` is used in
# dynamic_preferences_registry.py. The latter is parsed/invoked BEFORE django
# settings fills in langs with configula.
config = Configula()


def get_default_ocr_lang():
    """
    Returns default OCR language

    Pulls information by directly using Configula interface (as
    opposite to accessing settings.PAPERMERGE_OCR_DEFAULT_LANGUAGE
    """
    return config.get(
        'ocr',
        'default_language',
        default='deu'
    )


def get_ocr_langs(capitalize=True):
    """
    Returns a list of tuples as required by
    Django's choices ((key, value),(key, value), ...)

    Pulls information by directly using Configula interface (as
    opposite to accessing settings.PAPERMERGE_OCR_LANGUAGES
    """
    lang_dict = config.get(
        'ocr',
        'languages',
        default={
            'deu': 'Deutsch',
            'eng': 'English',
        }
    )

    return [
        (key, value.capitalize())
        for key, value in lang_dict.items()
    ]



from knox.models import AuthToken

from rest_framework_json_api import serializers


class TokenSerializer(serializers.ModelSerializer):
    """Serializer to list tokens"""
    token = serializers.CharField(max_length=256, required=False)

    class Meta:
        model = AuthToken
        resource_name = 'tokens'
        fields = ('token', 'digest', 'created', 'expiry')


class CreateTokenSerializer(serializers.Serializer):
    """Serializer to get input from user"""
    expiry_hours = serializers.IntegerField(default=744)



from django.dispatch import Signal

"""
Sent immediately after document upload complete.
Arguments:
    document_version - model instance of newly created document version
"""
document_post_upload = Signal()


"""
Sent immediately after (some of) document's page(s) were moved to folder.
Arguments:
    document_version - model instance of newly created document version
"""
page_move_to_folder = Signal()

"""
Sent immediately after (some of) document's page(s) were moved to another doc.
Arguments:
    document_version - model instance of newly created document version
"""
page_move_to_document = Signal()

"""
Sent immediately after document's page(s) were rotated.
Arguments:
    document_version - model instance of newly created document version
"""
page_rotate = Signal()

"""
Sent immediately after document's page(s) were deleted.
Arguments:
    document_version - model instance of newly created document version
"""
page_delete = Signal()


"""
Sent immediately after document's page(s) were reordered.
Arguments:
    document_version - model instance of newly created document version
"""
page_reorder = Signal()


# Sent by core.views.documents.create_folder
# Sent AFTER one single folder was created
"""
    providing_args=[
        "user_id",
        "level",
        "message",
        "folder_id"
    ]
"""
folder_created = Signal()

# Sent by core.views.nodes.nodes_view
# Sent AFTER one of more nodes were batch deleted
"""
    providing_args=[
        "user_id",
        "level",
        "message",
        # node tags is a list of
        # '<a href="{node.id}">{node.title}</a>'
        "node_tags"
    ]
"""
nodes_deleted = Signal()



import logging

from django.core.exceptions import ImproperlyConfigured
from django.apps import apps
from django.core.management import call_command

from celery import shared_task
from haystack.exceptions import NotHandled as IndexNotFoundException
from haystack import connections, connection_router


logger = logging.getLogger(__name__)


def split_identifier(identifier):
    """
    Break down the identifier representing the instance.
    Converts 'notes.note.23' into ('notes.note', 23).
    """
    bits = identifier.split('.')

    if len(bits) < 2:
        logger.error("Unable to parse object "
                     "identifer '%s'. Moving on..." % identifier)
        return (None, None)

    pk = bits[-1]
    # In case Django ever handles full paths...
    object_path = '.'.join(bits[:-1])
    return (object_path, pk)


def get_instance(model_class, pk):
    """
    Fetch the instance in a standarized way.
    """
    instance = None
    try:
        instance = model_class._default_manager.get(pk=pk)
    except model_class.DoesNotExist:
        logger.error("Couldn't load %s.%s.%s. Somehow it went missing?" %
                     (model_class._meta.app_label.lower(),
                      model_class._meta.object_name.lower(), pk))
    except model_class.MultipleObjectsReturned:
        logger.error("More than one object with pk %s. Oops?" % pk)
    return instance


def get_indexes(model_class):
    """
    Fetch the model's registered ``SearchIndex`` in a standarized way.
    """
    try:
        using_backends = connection_router.for_write(
            models=[model_class]
        )
        for using in using_backends:
            index_holder = connections[using].get_unified_index()
            yield index_holder.get_index(model_class), using
    except IndexNotFoundException:
        raise ImproperlyConfigured(
            "Couldn't find a SearchIndex for %s." % model_class
        )


def get_model_class(object_path):
    """
    Fetch the model's class in a standarized way.
    """
    bits = object_path.split('.')
    app_name = '.'.join(bits[:-1])
    classname = bits[-1]
    model_class = apps.get_model(app_name, classname)

    if model_class is None:
        raise ImproperlyConfigured("Could not load model '%s'." %
                                   object_path)
    return model_class


@shared_task
def update_index():
    call_command("update_index")



import logging
from django.core.management.base import BaseCommand

from papermerge.core.models import (
    BaseTreeNode,
    Access
)
from papermerge.core.auth import (
    create_access_perms
)

logger = logging.getLogger(__name__)


class Command(BaseCommand):

    help = """Lists/Updates Access Models associated with nodes.
"""

    def add_arguments(self, parser):
        parser.add_argument(
            '--count',
            '-c',
            action="store_true",
            help="Count nodes with/without associated access model."
        )
        parser.add_argument(
            '--update',
            '-u',
            action="store_true",
            help="Updated nodes without associated access model."
        )

    def run_count(
        self,
    ):
        total_count = BaseTreeNode.objects.count()
        without_access_count = BaseTreeNode.objects.filter(
            access__isnull=True
        ).count()

        print(
            f"total={total_count}, without_access={without_access_count}"
        )

    def run_update(
        self
    ):
        perms = create_access_perms()

        for node in BaseTreeNode.objects.all():
            if node.access_set.count() == 0:
                access = Access.objects.create(
                    user=node.user,
                    access_type='allow',
                    node=node
                )
                access.permissions.add(*perms)

    def handle(self, *args, **options):
        count = options.get(
            'count',
            False
        )
        update = options.get(
            'update',
            False
        )
        if count:
            self.run_count()
        elif update:
            self.run_update()



import logging

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer

from papermerge.core.models import Document
from papermerge.core.models import utils as model_utils
from papermerge.notifications.mixins import RequireAuth

logger = logging.getLogger(__name__)


def _update_document_ocr_status(event: dict, ocr_status):
    """
    Example of event:
    {'type': 'ocrdocumenttask.taskreceived', 'document_id': 22, 'lang': 'deu'}
    """
    if 'document_id' in event:
        document_id = event['document_id']
        try:
            document = Document.objects.get(pk=document_id)
            document.ocr_status = ocr_status
            document.save()
        except Document.DoesNotExist as exc:
            # not end of the world, but still good to know
            logger.warning(
                f"Consumer did not found the document_id={document_id}"
            )
            logger.warning(exc)
            # life goes on...


class DocumentConsumer(RequireAuth, JsonWebsocketConsumer):
    """
    Synchronous Consumer.

    It is safe to run Django ORM operations inside synchronous consumer.
    """

    group_name = "ocr_document_task"

    def disconnect(self, close_code):
        logger.debug("DISCONNECT")

        async_to_sync(
            self.channel_layer.group_discard
        )(self.group_name, self.channel_name)

    def ocrdocumenttask_taskreceived(self, event: dict):
        logger.debug(
            "ocrdocumenttask_taskreceived triggered"
        )
        self._ocr_document_event(
            event=event,
            type=model_utils.OCR_STATUS_RECEIVED
        )

    def ocrdocumenttask_taskstarted(self, event: dict):
        logger.debug(
            "ocrdocumenttask_taskstarted triggered"
        )
        self._ocr_document_event(
            event=event,
            type=model_utils.OCR_STATUS_STARTED
        )

    def ocrdocumenttask_tasksucceeded(self, event: dict):
        logger.debug(
            "ocrdocumenttask_tasksucceeded triggered"
        )
        self._ocr_document_event(
            event=event,
            type=model_utils.OCR_STATUS_SUCCEEDED
        )

    def ocrdocumenttask_taskfailed(self, event: dict):
        logger.debug(
            "ocrdocumenttask_taskfailed triggered"
        )
        self._ocr_document_event(
            event=event,
            type=model_utils.OCR_STATUS_FAILED
        )

    def _ocr_document_event(self, event, type):
        if event['user_id'] != str(self.user.id):
            # notification is intended only for user who initiated it
            return

        _update_document_ocr_status(event, type)

        logger.debug(
            f"Doc Consumer ev={event} type={type} for user_id={self.user.id}"
        )
        self.send_json(event)



# Generated by Django 4.0.8 on 2022-12-22 06:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0003_alter_basetreenode_polymorphic_ctype_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='automate',
            name='dst_folder',
        ),
        migrations.RemoveField(
            model_name='automate',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='automate',
            name='user',
        ),
        migrations.RemoveField(
            model_name='kvstorecompitem',
            name='comp_item',
        ),
#        migrations.RemoveField(
#            model_name='kvstorecompitem',
#            name='kvstore_ptr',
#        ),
        migrations.RemoveField(
            model_name='kvstorecompnode',
            name='node',
        ),
        migrations.RemoveField(
            model_name='kvstorenode',
            name='kvstore_ptr',
        ),
        migrations.RemoveField(
            model_name='kvstorenode',
            name='node',
        ),
        migrations.RemoveField(
            model_name='kvstorepage',
            name='kvstore_ptr',
        ),
        migrations.RemoveField(
            model_name='kvstorepage',
            name='page',
        ),
        migrations.DeleteModel(
            name='Access',
        ),
        migrations.DeleteModel(
            name='Automate',
        ),
        migrations.DeleteModel(
            name='KVStore',
        ),
        migrations.DeleteModel(
            name='KVStoreCompItem',
        ),
        migrations.DeleteModel(
            name='KVStoreCompNode',
        ),
        migrations.DeleteModel(
            name='KVStoreNode',
        ),
        migrations.DeleteModel(
            name='KVStorePage',
        ),
    ]



import logging

from rest_framework_json_api.views import ModelViewSet

from papermerge.core.serializers import TagSerializer
from papermerge.core.models import Tag

from .mixins import RequireAuthMixin

logger = logging.getLogger(__name__)


class TagsViewSet(RequireAuthMixin, ModelViewSet):
    serializer_class = TagSerializer

    def get_queryset(self, *args, **kwargs):
        # This is workaround warning issued when runnnig
        # `./manage.py generateschema`
        # https://github.com/carltongibson/django-filter/issues/966
        if not self.request:
            return Tag.objects.none()

        return Tag.objects.filter(user=self.request.user)

    def perform_create(self, serializer):
        serializer.save(user_id=self.request.user.id)



import logging

from django.dispatch import receiver
from django.db.models.signals import post_save, post_delete

from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync

from papermerge.core.models import (
    Document,
    Folder
)

logger = logging.getLogger(__name__)


@receiver([post_delete, post_save], sender=Document)
@receiver([post_delete, post_save], sender=Folder)
def if_inbox_then_refresh(sender, instance, **kwargs):
    """
    Inform inbox_refresh channel group that user's inbox was updated
    """
    # Folder or Document instance was deleted/moved from//to user's Inbox folder
    try:
        instance.refresh_from_db()
    except (Document.DoesNotExist, Folder.DoesNotExist):
        logger.warning('Too late - Document/Folder was already deleted')
        return

    try:
        if instance.parent and instance.parent.title == Folder.INBOX_TITLE:
            channel_layer = get_channel_layer()
            async_to_sync(channel_layer.group_send)(
                "inbox_refresh",
                {"type": "inbox.refresh", "user_id": str(instance.user.pk)}
            )
    except Exception as ex:
        logger.warning(ex, exc_info=True)



import logging
import datetime
import os

from django.core.management import BaseCommand

from papermerge.core.backup_restore import backup_documents
from papermerge.core.models import User

logger = logging.getLogger(__name__)


def list_users_and_exit():
    print("id\tusername\temail")
    print("----------------------------------")
    for user in User.objects.all():
        print(f"{user.id}\t{user.username}\t{user.email}")

    return True


class Command(BaseCommand):
    help = """
        Backup all documents and their folder structure to an archive.
        If --user/-u is specified - will backup all documents of
        the specific user.

        If --user/u is NOT specified - will backup documents of all users.
        --include-user-password will include user's password digest into backup
        archive
    """

    def add_arguments(self, parser):
        parser.add_argument(
            'location',
            nargs='?',
            type=str,
            help="A directory or a file name (tar archive)."
            " If provided file is missing - will create one."
        )
        # Papermerge is multi-user system. There may be different users
        # with diferent folder structure each. Perform backup on specified one.
        parser.add_argument(
            "-u",
            "--user",
            help="""
            username of the user to perform backup for. If not given
            will perform backup for all users.
            """
        )
        parser.add_argument(
            "-p",
            "--include-user-password",
            action="store_true",
            # by default do not include user password's digest
            default=False,
            help="""
            Include user's password field (digest of user's password) into
            backup archive.
            """
        )
        parser.add_argument(
            "-l",
            "--list-users",
            action='store_true',
            help="List exiting users and quit."
        )

    def handle(self, *args, **options):

        date_string = datetime.datetime.now().strftime("%d_%m_%Y-%H_%M_%S")
        default_filename = f"backup_{date_string}.tar"

        username = options.get('user')
        location = options.get('location') or default_filename
        include_user_password = options.get('include_user_password')

        if options.get('list_users'):
            list_users_and_exit()
            return

        user = None
        if username:
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                logger.error(
                    f"Username {username} not found."
                )
                return

        # consider the case when user provides directory location i.e.
        # ./manage.py backup /backup/papermerge/

        if os.path.isdir(location):
            # in case location is a non existing directory path e.g. "blah/"
            # os.path.isdir will return False
            backup_location = os.path.join(location, default_filename)
        else:
            backup_location = location

        try:
            with open(backup_location, 'wb') as backup_file:
                backup_documents(
                    backup_file=backup_file,
                    user=user,
                    include_user_password=include_user_password
                )
        except IsADirectoryError:
            logger.error(
                "Provided location is a directory which does not exist."
                "Please create it first and try again."
            )



import time
import logging
import re
import functools
from datetime import datetime

from django.conf import settings
from django.utils.html import format_html
from django.urls import reverse


logger = logging.getLogger(__name__)


def date_2int(kv_format, str_value):
    # maps PAPERMERGE_METADATA_DATE_FORMATS to
    # https://docs.python.org/3.8/library/datetime.html#strftime-and-strptime-format-codes

    if not str_value:
        return 0

    format_map = {
        'dd.mm.yy': '%d.%m.%y',
        'dd.mm.yyyy': '%d.%m.%Y',
        'dd.M.yyyy': '%d.%B.%Y',
        'month': '%B'
    }
    try:
        _date_instance = datetime.strptime(
            str_value, format_map[kv_format]
        )
    except Exception as e:
        # this is expected because of automated
        # extraction of metadata may fail.
        logger.debug(
            f"While converting date user format {e}"
        )
        return 0

    return _date_instance.timestamp()


def money_2int(kv_format, str_value):
    return number_2int(kv_format, str_value)


def number_2int(kv_format, str_value):
    """
    kv_format for number is usually something like this:

        dddd
        d,ddd
        d.ddd

    So converting to an integer means just remove from string
    non-numeric characters and cast remaining str to integer.
    """
    if str_value:
        line = re.sub(r'[\,\.]', '', str_value)
        return line

    return 0


def node_tag(node):

    node_url = reverse("core:node", args=(node.id,))
    tag = format_html(
        "<a href='{}'>{}</a>",
        node_url,
        node.title
    )

    return tag


def document_tag(node):

    node_url = reverse("core:document", args=(node.id,))
    tag = format_html(
        "<a href='{}'>{}</a>",
        node_url,
        node.title
    )

    return tag


class Timer:
    """
    Timer class used to measure how much time
    certain code block took to complete.

    Example:

        with Timer() as t:
            main_ocr_page(...)

        logger.info(
            f"OCR took {t:.2f} seconds to complete"
        )
    """

    def __init__(self):
        self.total = None

    def __enter__(self):
        self.start = time.time()
        # important, because 'as' variable
        # is assigned only the result of __enter__()
        return self

    def __exit__(self, type, value, traceback):
        self.end = time.time()
        self.total = self.end - self.start

    def __str__(self):
        return f"{self.total:.2f}"


def filter_node_id(value):
    """Invalid values of node id will be
    filtered out (return None).

    Valid values for node id will pass
    and will be returned as integers.
    """
    if not value:
        return None

    if isinstance(value, str):
        if value.isnumeric():
            return int(value)
        return None

    if isinstance(value, int):
        if value < 0:
            return None

        return value

    return None


def remove_backup_filename_id(value: str) -> str:
    """
    value is a string that looks like something__number,
    i.e. consists of two parts separated by double underscore.
    Second part (__number) is a number.
    Examples:

        blah.pdf__23
        boo__1
        asdlaksd__100

    This function returns first part of the string:

    value: blah.pdf__23 => result: blah.pdf
           boo__1  => boo

    Other examples:

        boox_1       => boox
        boox         => boox
        boox_____100 => boox
        None         => None
    """
    # works only with string input
    if not value:
        return None

    if not isinstance(value, str):
        return value

    result = value.split('_')

    if len(result) <= 2:
        return result[0]

    return "_".join(result[0:-2])


def namespaced(name):
    if settings.PAPERMERGE_NAMESPACE is not None:
        return f"{settings.PAPERMERGE_NAMESPACE}__{name}"

    return name


def clock(func):
    """
    Logs time (in seconds) it take the decorated function to execute
    """

    @functools.wraps(func)
    def inner(*args, **kwargs):
        t0 = time.perf_counter()

        result = func(*args, **kwargs)

        elapsed = time.perf_counter() - t0
        name = func.__name__
        arg_lst = [repr(arg) for arg in args]
        arg_lst.extend(f'{k}={v!r}' for k, v in kwargs.items())
        arg_str = ','.join(arg_lst)

        logger.debug(f'{elapsed:0.6f}s {name} called with {arg_str}')

        return result

    return inner



import logging

from rest_framework.renderers import JSONRenderer as rest_framework_JSONRenderer
from rest_framework.parsers import JSONParser as rest_framework_JSONParser
from rest_framework.generics import GenericAPIView
from rest_framework import status
from rest_framework.response import Response
from rest_framework.parsers import FileUploadParser
from rest_framework_json_api.views import ModelViewSet
from rest_framework_json_api.renderers import JSONRenderer
from drf_spectacular.utils import (
    extend_schema,
    OpenApiParameter,
    OpenApiExample
)

from papermerge.core.serializers.document import Data_DocumentDetailsSerializer

from papermerge.core.serializers import (
    DocumentDetailsSerializer,
    DocumentsMergeSerializer,
    DocumentVersionOcrTextSerializer
)
from papermerge.core.models import Document
from papermerge.core.exceptions import APIBadRequest

from .mixins import RequireAuthMixin
from .utils import total_merge


logger = logging.getLogger(__name__)


class DocumentUploadView(RequireAuthMixin, GenericAPIView):
    parser_classes = [FileUploadParser]
    serializer_class = DocumentDetailsSerializer
    renderer_classes = (JSONRenderer,)
    http_method_names = ["put"]

    @extend_schema(
        operation_id="Upload file",
        parameters=[
            OpenApiParameter(
                name='Content-Disposition',
                type=str,
                location=OpenApiParameter.HEADER,
                required=True,
                response=False,
                examples=[
                    OpenApiExample(
                        name="Content disposition header for file named"
                        "mydoc_1.pdf",
                        value="attachment; filename=mydoc_1.pdf"
                    ),
                    OpenApiExample(
                        name="Content disposition header for file"
                        "named abc.pdf",
                        value="attachment; filename=abc.pdf"
                    )
                ]
            ),
        ],
        request={
            'application/octet-stream': bytes
        },
        responses={201: Data_DocumentDetailsSerializer},
    )
    def put(self, request, document_id, file_name):
        """
        Uploads a file for given document node. If last version of the
        document does not have any file associated yet, this REST API call
         will associated given file with document’s last version. If, on
        the other hand, last version of the document already has a file
        associated with it - a new document version will be created and
        associated it with respective file.

        Request body should contain file data. Please note that you need to
        specify ``Content-Disposition`` header with value
        'attachment; filename={file_name}'.
        """
        payload = request.data['file']

        doc = Document.objects.get(pk=document_id)
        doc.upload(
            payload=payload,
            file_path=payload.temporary_file_path(),
            file_name=file_name
        )
        serializer = DocumentDetailsSerializer(instance=doc)

        return Response(serializer.data, status=status.HTTP_201_CREATED)


class DocumentOcrTextView(RequireAuthMixin, GenericAPIView):
    serializer_class = DocumentVersionOcrTextSerializer
    parser_classes = (rest_framework_JSONParser,)
    renderer_classes = (rest_framework_JSONRenderer,)
    queryset = Document.objects.all()

    @extend_schema(
        operation_id="Document OCR Text",
        parameters=[
            OpenApiParameter(
                name='page_numbers[]',
                description=(
                    "Filter pages by provided page numbers"
                ),
                required=False,
                type={'type': 'array', 'items': {'type': 'number'}}
            ),
            OpenApiParameter(
                name='page_ids[]',
                description=(
                    "Filter pages by provided page ids"
                ),
                required=False,
                type={'type': 'array', 'items': {'type': 'string'}}
            ),
        ]
    )
    def get(self, request, pk, *args, **kwargs):
        """Retrieve OCRed text of the document

        You can filter pages for which OCRed text is to be received either by
        page numbers or by page ids. When both filters are empty - retrieve
        OCRed text of the whole document (i.e. of its last document version)
        """

        # Document instance
        instance = self.get_object()
        document_version = instance.versions.last()
        # For what page number does user want to get OCR text ?
        # If page_numbers parameter is empty - get OCR text for all pages
        # of the document version
        try:
            page_numbers = self.request.GET.getlist('page_numbers[]', [])
            page_numbers = [int(number) for number in page_numbers]
        except ValueError:
            page_numbers = []

        page_ids = self.request.GET.getlist('page_ids[]', [])

        text = document_version.get_ocred_text(
            page_numbers=page_numbers,
            page_ids=page_ids
        )
        serializer = self.get_serializer(data={'text': text})

        if serializer.is_valid():
            return Response(data=serializer.data)

        return Response(
            serializer.errors,
            status=status.HTTP_400_BAD_REQUEST
        )


class DocumentsMergeView(RequireAuthMixin, GenericAPIView):
    serializer_class = DocumentsMergeSerializer
    parser_classes = (rest_framework_JSONParser,)
    renderer_classes = (rest_framework_JSONRenderer,)

    @extend_schema(operation_id="Documents Merge")
    def post(self, request):
        """Merge source document into destination

        A new document version is created on the target (destination document)
        from all pages of the last version of the source.
        **Source document is deleted**.

        Merge operation is useful when you have same document scanned several
        times. Instead of keeping two similar scanned copies of the same
        document, better to merge them as two versions of the same document.
        For example, say you have two scans of the same document A_lq.pdf and
        A_hq.pdf where A_lq.pdf low quality scan and A_hq.pdf is high quality
        scan. You want to merge A_lq.pdf and A_hq.pdf document into one so
        that higher quality scan will be lastest version.
        In this case you need to set A_hq.pdf as source and A_lq.pdf as
        destination.

        Notice that OCR data (if any) is reused, this means that after
        merge operation you don't have to re-run OCR as the OCR data of the
        source document is reuse/copied to the target.
        """
        serializer = self.serializer_class(data=request.data)

        if serializer.is_valid():
            _merge_documents(
                src_uuid=serializer.data['src'],
                dst_uuid=serializer.data['dst']
            )
            return Response(
                data=serializer.data,
                status=status.HTTP_204_NO_CONTENT
            )

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


class DocumentDetailsViewSet(RequireAuthMixin, ModelViewSet):
    """
    Document details endpoint.
    """
    serializer_class = DocumentDetailsSerializer
    queryset = Document.objects.all()
    renderer_classes = (JSONRenderer,)

    http_method_names = [
        "get",
        "delete",
        "patch",
        "head",
        "options"
    ]


def _merge_documents(src_uuid, dst_uuid):
    try:
        src = Document.objects.get(pk=src_uuid)
    except Document.DoesNotExists:
        raise APIBadRequest(f"src={src_uuid} not found")

    try:
        dst = Document.objects.get(pk=dst_uuid)
    except Document.DoesNotExists:
        raise APIBadRequest(f"dst={dst_uuid} not found")

    src_version = src.versions.last()
    dst_new_version = dst.version_bump(
        page_count=src_version.pages.count(),
        short_description='document merge'
    )
    total_merge(
        src_old_version=src_version,
        dst_new_version=dst_new_version
    )



from django.urls import re_path

from .consumers import document as doc_consumer
from .consumers import inbox_refresh as inbox_refresh_consumer
from .consumers import DefaultConsumer

websocket_urlpatterns = [
    re_path(
        r'ws/document/$',
        doc_consumer.DocumentConsumer.as_asgi()
    ),
    re_path(
        r'ws/nodes/inbox-refresh/$',
        inbox_refresh_consumer.InboxRefreshConsumer.as_asgi()
    ),
    re_path(
        r'ws/',
        DefaultConsumer.as_asgi()
    ),
]



# Generated by Django 3.2.13 on 2022-06-19 08:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='user',
            name='role',
        ),
        migrations.DeleteModel(
            name='Role',
        ),
    ]



from collections import OrderedDict

from rest_framework.validators import UniqueTogetherValidator
from rest_framework_json_api import serializers
from rest_framework_json_api.relations import ResourceRelatedField
from rest_framework_json_api.utils import get_resource_type_from_instance

from papermerge.core.models import Folder
from .tag import ColoredTagListSerializerField


class NodeRelatedField(ResourceRelatedField):

    def to_representation(self, value):
        if getattr(self, "pk_field", None) is not None:
            pk = self.pk_field.to_representation(value.pk)
        else:
            pk = value.pk

        resource_type = self.get_resource_type_from_included_serializer()
        if resource_type is None or not self._skip_polymorphic_optimization:
            resource_type = get_resource_type_from_instance(value)

        attributes = OrderedDict([
            ("title", value.title)
        ])

        return OrderedDict([
            ("type", resource_type),
            ("id", str(pk)),
            ("attributes", attributes)
        ])


class FolderSerializer(serializers.ModelSerializer):

    parent = ResourceRelatedField(queryset=Folder.objects)
    tags = ColoredTagListSerializerField(required=False)

    class Meta:
        model = Folder
        resource_name = 'folders'
        fields = (
            'id',
            'title',
            'parent',
            'tags',
            'created_at',
            'updated_at'
        )
        validators = [
            UniqueTogetherValidator(
                queryset=Folder.objects.all(),
                fields=['parent', 'title']
            )
        ]



import logging
from django.utils.module_loading import import_string
from django.conf import settings as django_settings

from .app_settings import settings


logger = logging.getLogger(__name__)


def get_storage_class(import_path=None):
    return import_string(
        import_path or settings.DEFAULT_FILE_STORAGE
    )


def get_storage_instance():
    storage_klass = get_storage_class()
    storage_kwargs = settings.FILE_STORAGE_KWARGS or {}
    storage_kwargs['location'] = django_settings.MEDIA_ROOT

    return storage_klass(**storage_kwargs)


def abs_path(some_relative_path):
    storage_instance = get_storage_instance()

    return storage_instance.abspath(some_relative_path)


# TODO: remove this code
storage_class = get_storage_class()

# TODO: remove this code
storage_kwargs = settings.FILE_STORAGE_KWARGS or {}
# This line will return '' if code runs before MEDIA_ROOT is being set
# e.g. when preforking in uwsgi
# Remove this code and access default storage only via `get_storage_instance`
storage_kwargs['location'] = django_settings.MEDIA_ROOT

# TODO: remove this code
default_storage = storage_class(**storage_kwargs)



from django.urls import reverse

from rest_framework.serializers import BaseSerializer
from rest_framework_json_api import serializers
from drf_spectacular.openapi import OpenApiTypes
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers as rest_serializers
from papermerge.core.models import DocumentVersion


class DocumentVersionSerializer(serializers.ModelSerializer):

    download_url = serializers.SerializerMethodField()

    class Meta:
        model = DocumentVersion
        resource_name = 'document-versions'
        fields = (
            'id',
            'number',
            'lang',
            'file_name',
            'pages',
            'size',
            'page_count',
            'short_description',
            'document',
            'download_url'
        )

    @extend_schema_field(OpenApiTypes.STR)
    def get_download_url(self, obj):
        return reverse('download-document-version', args=[str(obj.pk)])


class DocumentVersionOcrTextSerializer(rest_serializers.Serializer):
    """Returns OCRed Text of the document"""
    text = serializers.CharField(required=False, allow_blank=True)


class DocumentVersionDownloadSerializer(BaseSerializer):
    """DocumentVersion => corresponding file bytes"""

    @property
    def fields(self):
        return {}

    def to_representation(self, instance):
        file_abs_path = instance.abs_file_path()

        with open(file_abs_path, "rb") as file:
            content = file.read()

        return content



from rest_framework_json_api import serializers

from papermerge.core.serializers.tag import ColoredTagListSerializerField


class SearchResultSerializer(serializers.Serializer):
    id = serializers.UUIDField()
    text = serializers.CharField(required=False, default='')
    title = serializers.CharField()
    tags = ColoredTagListSerializerField(required=False)
    highlight = serializers.CharField()
    breadcrumb = serializers.ListField(
        child=serializers.CharField()
    )
    node_type = serializers.ChoiceField(choices=['document', 'folder'])
    user_id = serializers.UUIDField()



from rest_framework.serializers import Serializer
from rest_framework.validators import UniqueTogetherValidator

from rest_framework_json_api import serializers
from rest_framework_json_api.relations import ResourceRelatedField

from papermerge.core.models import Folder
from papermerge.core.models import Document

from .document_version import DocumentVersionSerializer
from .tag import ColoredTagListSerializerField


class CurrentUserDefaultLang:
    """
    Returns the current user's default OCR language.
    """
    requires_context = True

    def __call__(self, serializer_field):
        user = serializer_field.context['request'].user
        return user.preferences['ocr__language']


class DocumentSerializer(serializers.ModelSerializer):
    size = serializers.IntegerField(required=False)
    page_count = serializers.IntegerField(required=False)
    parent = ResourceRelatedField(queryset=Folder.objects)
    file_name = serializers.CharField(required=False)
    ocr = serializers.BooleanField(required=False)
    lang = serializers.CharField(
        required=False,
        default=CurrentUserDefaultLang()
    )
    ocr_status = serializers.CharField(required=False)
    tags = ColoredTagListSerializerField(required=False)

    class Meta:
        model = Document
        resource_name = 'documents'
        fields = (
            'id',
            'title',
            'lang',
            'ocr',
            'ocr_status',
            'file_name',
            'parent',
            'tags',
            'size',
            'page_count',
            'created_at',
            'updated_at'
        )
        validators = [
            UniqueTogetherValidator(
                queryset=Document.objects.all(),
                fields=['parent', 'title']
            )
        ]

    def create(self, validated_data):
        return Document.objects.create_document(
            size=0,
            page_count=0,
            file_name=validated_data['title'],
            **validated_data
        )


class DocumentDetailsSerializer(serializers.ModelSerializer):
    parent = ResourceRelatedField(queryset=Folder.objects)
    versions = DocumentVersionSerializer(many=True, read_only=True)

    class Meta:
        model = Document
        resource_name = 'documents'
        fields = (
            'id',
            'title',
            'lang',
            'ocr',
            'ocr_status',
            'parent',
            'versions',
            'created_at',
            'updated_at'
        )


class Data_DocumentDetailsSerializer(Serializer):
    data = DocumentDetailsSerializer()


class DocumentsMergeSerializer(serializers.Serializer):
    src = serializers.UUIDField()
    dst = serializers.UUIDField()



from drf_spectacular.utils import extend_schema

from django.db import transaction

from rest_framework.decorators import action

from dynamic_preferences.users.viewsets import UserPreferencesViewSet

from papermerge.core.serializers import CustomUserPreferenceSerializer
from papermerge.core.auth import CustomModelPermissions


OPENAPI_PREFERENCE = {
    'type': 'object',
    'properties': {
        'ocr__language': {
            'type': 'string'
        },
        'ocr__trigger': {
            'type': 'string',
            'enum': ['auto', 'manual']
        },
        'localization__data_format': {
            'type': 'string'
        },
        'localization__time_format': {
            'type': 'string'
        },
        'localization__timezone': {
            'type': 'string'
        },
    }
}


class CustomUserPreferencesViewSet(UserPreferencesViewSet):
    serializer_class = CustomUserPreferenceSerializer
    permission_classes = [CustomModelPermissions]

    @extend_schema(
        responses={
            200: {
                'type': 'object',
                'properties': {
                    'data': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'additional_data': {
                                    'type': 'object',
                                    'properties': {
                                        'choices': {
                                            'type': 'array'
                                        }
                                    }
                                },
                                'default': {
                                    'type': 'string'
                                },
                                'field': {
                                    'type': 'object',
                                    'properties': {
                                        'class': {
                                            'type': 'string'
                                        },
                                        'input_type': {
                                            'type': 'string'
                                        }
                                    }
                                },
                                'help_text': {
                                    'type': 'string'
                                },
                                'id': {
                                    'type': 'string'
                                },
                                'identifier': {
                                    'type': 'string'
                                },
                                'name': {
                                    'type': 'string'
                                },
                                'section': {
                                    'type': 'string'
                                },
                                'value': {
                                    'type': 'string'
                                },
                            }
                        }
                    }
                }
            },
        },
        request={
            'application/vnd.api+json': OPENAPI_PREFERENCE,
            'application/json': OPENAPI_PREFERENCE
        }
    )
    @action(detail=False, methods=["post"])
    @transaction.atomic
    def bulk(self, request, *args, **kwargs):
        return super().bulk(request, *args, **kwargs)



from papermerge.core.models import Folder
from rest_framework_json_api.views import ModelViewSet
from rest_framework_json_api.renderers import JSONRenderer

from papermerge.core.serializers import FolderSerializer
from .mixins import RequireAuthMixin


class FoldersViewSet(RequireAuthMixin, ModelViewSet):
    """
    Folders endpoint
    """
    queryset = Folder.objects.all()
    serializer_class = FolderSerializer
    renderer_classes = (JSONRenderer,)

    def perform_create(self, serializer):
        serializer.save(user_id=self.request.user.pk)



import os
import uuid
import logging

from django.db import models
from django.utils.translation import gettext_lazy as _
from papermerge.core.storage import abs_path

from pdf2image import convert_from_path
from pdf2image.generators import counter_generator

from papermerge.core.lib.path import DocumentPath


logger = logging.getLogger(__name__)


class DocumentVersion(models.Model):
    """Document Version

    Document can have one or multiple versions.
    Document has at least one version associated (the original).
    Each document version has a number - which starts with 1 (one) i.e.
    original document version is - document version 1 (one).
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)

    document = models.ForeignKey(
        to='Document',
        on_delete=models.CASCADE,
        related_name='versions',
        verbose_name=_('Document')
    )

    lang = models.CharField(
        _('Language'),
        max_length=8,
        blank=False,
        null=False,
        default='deu'
    )
    # version number
    number = models.IntegerField(
        default=1,  # Document versioning starts with 1
        verbose_name=_('Version number')
    )
    #: basename + ext of uploaded file.
    #: other path details are deducted from user_id and document_id
    file_name = models.CharField(
        max_length=1024,
        blank=True,
        null=True,
    )
    size = models.BigIntegerField(
        help_text="Size of file_orig attached. Size is in Bytes",
        blank=False,
        null=False,
        default=0
    )
    page_count = models.IntegerField(
        blank=False,
        default=0
    )
    short_description = models.CharField(
        max_length=128,
        blank=True,
        default=''
    )

    text = models.TextField(blank=True)

    class Meta:
        ordering = ('number',)
        verbose_name = _('Document version')
        verbose_name_plural = _('Document versions')

    def __repr__(self):
        return f"DocumentVersion(id={self.pk}, number={self.number})"

    def abs_file_path(self):
        return abs_path(
            self.document_path.url
        )

    def generate_previews(self, page_number=None):
        logger.debug('generate_previews BEGIN')
        abs_dirname = abs_path(self.document_path.dirname_sidecars())

        kwargs = {
            'pdf_path': abs_path(self.document_path.url),
            'output_folder':  abs_dirname,
            'fmt': 'jpg',
            'size': (900,),
            'output_file': counter_generator(padding_goal=3)
        }

        # in case page_number not None - generate only specific
        # page number's preview
        if page_number:
            kwargs['first_page'] = page_number
            kwargs['last_page'] = page_number

        os.makedirs(
            abs_dirname,
            exist_ok=True
        )
        # generates jpeg previews of PDF file using pdftoppm (poppler-utils)
        convert_from_path(**kwargs)
        logger.debug('generate_previews END')

    @property
    def is_archived(self):
        """
        Returns True if document version is archived.

        Document version is considered archived if
        it is non-last version of the document.
        Only last document version is editable, or to put
        in other words - archived document versions are not
        editable.
        """
        return self != self.document.versions.last()

    @property
    def document_path(self):
        return DocumentPath(
            user_id=self.document.user.pk,
            document_id=self.document.pk,
            version=self.number,
            file_name=self.file_name,
        )

    def create_pages(self, page_count=None):
        """
        Creates Page models for current document version.

        If no argument is supplied, will read
        number of pages from `self.page_count`
        """

        new_page_count = self.page_count  # may be zero

        if page_count:
            # provided non empty argument overrides `self.page_count`
            new_page_count = page_count

        if not new_page_count:
            # Number of pages in document version is 0 or None
            # Also no argument was supplied. Nothing to do.
            return

        for page_number in range(1, new_page_count + 1):
            self.pages.create(
                number=page_number,
                page_count=new_page_count,
                lang=self.lang
            )

        if new_page_count and new_page_count != self.page_count:
            self.page_count = new_page_count
            self.save()

    @property
    def has_combined_text(self):
        """
        Returns ``True`` if at least one of document versions'
        page contains non empty, non whitespace text - otherwise
        returns ``False``
        """
        text = ''

        for page in self.pages.all():
            text = text + ' ' + page.text

        return len(text.strip()) != 0

    def update_text_field(self, streams):
        """Update document versions's text field from IO streams.

        Arguments:
            ``streams`` - a list of IO text streams

        It will update text field of all associated pages first
        and then concatinate all text field into doc.text field.

        Returns True if document version contains non empty non whitespace
        text (i.e it was OCRed)
        """
        text = []

        logger.debug(
            'document.update_text_field: '
            f'document_id={self.pk} streams_count={len(streams)}'
        )

        for page, stream in zip(self.pages.all(), streams):
            if len(page.text) == 0:
                page.update_text_field(stream)
                text.append(page.stripped_text)

        stripped_text = ' '.join(text)
        stripped_text = stripped_text.strip()
        if stripped_text:
            self.text = stripped_text
            self.save()

        return self.has_combined_text

    def get_ocred_text(
        self,
        page_numbers: list = (),
        page_ids: list = ()
    ) -> str:
        """
        Returns OCRed text of given pages.

        You can filter pages for which OCRed is requested either be page numbers
        or by page_ids.
        If both page_numbers and page_ids are empty i.e. no filters, then
        return `self.text`.
        """
        pages_text = " ".join([
            page.text for page in self.pages.all()
            if page.number in page_numbers or str(page.pk) in page_ids
        ])

        if page_ids or page_numbers:
            result = pages_text.strip()
        else:
            # when both filters are empty, return the `self.text` field
            result = self.text.strip()

        return result



from rest_framework_json_api import serializers
from taggit.serializers import TagListSerializerField, TagList

from papermerge.core.models import Tag


class TagSerializer(serializers.ModelSerializer):

    class Meta:
        model = Tag
        resource_name = 'tags'
        fields = (
            'id',
            'name',
            'bg_color',
            'fg_color',
            'description',
            'pinned',
        )

    def save(self, **attrs):
        user_id = attrs.get('user_id', None)
        name = self.validated_data['name']
        try:
            Tag.objects.get(name=name, user_id=user_id)
        except Tag.DoesNotExist:
            pass
        else:
            raise serializers.ValidationError(
                f"Tag with name='{name}' already exists"
            )

        return super().save(**attrs)


class ColoredTagListSerializerField(TagListSerializerField):
    child = TagSerializer()

    class Meta:
        fields = '__all__'

    def to_representation(self, value):
        if not isinstance(value, TagList):
            if not isinstance(value, list):
                # value may come in as elastic search attrlist
                # or value may come in as Django's ORM related manager
                all_method = getattr(value, "all", None)
                if callable(all_method):
                    if self.order_by:
                        tags = value.all().order_by(*self.order_by)
                    else:
                        tags = value.all()
                else:
                    # arrived here via elastic search which returns
                    # its own iterator
                    tags = value
                value = [
                    {
                        'name': tag.name,
                        'bg_color': tag.bg_color,
                        'fg_color': tag.fg_color
                    } for tag in tags
                ]
            value = TagList(value, pretty_print=self.pretty_print)

        return value



"""
OpenAPI components appended during openapi schema generation.

Learn more about OpenAPI components:
    https://swagger.io/docs/specification/components/
"""

JSONAPI_COMPONENTS = {
    "schemas": {
        "jsonapi": {
            "type": "object",
            "description": "The server's implementation",
            "properties": {
                "version": {"type": "string"},
                "meta": {"$ref": "#/components/schemas/meta"},
            },
            "additionalProperties": False,
        },
        "resource": {
            "type": "object",
            "required": ["type", "id"],
            "additionalProperties": False,
            "properties": {
                "type": {"$ref": "#/components/schemas/type"},
                "id": {"$ref": "#/components/schemas/id"},
                "attributes": {
                    "type": "object",
                    # ...
                },
                "relationships": {
                    "type": "object",
                    # ...
                },
                "links": {"$ref": "#/components/schemas/links"},
                "meta": {"$ref": "#/components/schemas/meta"},
            },
        },
        "link": {
            "oneOf": [
                {
                    "description": "a string containing the link's URL",
                    "type": "string",
                    "format": "uri-reference",
                },
                {
                    "type": "object",
                    "required": ["href"],
                    "properties": {
                        "href": {
                            "description": "a string containing the link's URL",
                            "type": "string",
                            "format": "uri-reference",
                        },
                        "meta": {"$ref": "#/components/schemas/meta"},
                    },
                },
            ]
        },
        "links": {
            "type": "object",
            "additionalProperties": {"$ref": "#/components/schemas/link"},
        },
        "reltoone": {
            "description": "a singular 'to-one' relationship",
            "type": "object",
            "properties": {
                "data": {"$ref": "#/components/schemas/relationshipToOne"},
            },
        },
        "relationshipToOne": {
            "description": "reference to other resource in a to-one relationship",  # noqa
            "type": "object",
            "properties": {
                "type": {"type": "string"},
                "id": {"type": "string", "format": "uuid"}
            },
        },
        "reltomany": {
            "description": "a multiple 'to-many' relationship",
            "type": "object",
            "properties": {
                "data": {"$ref": "#/components/schemas/relationshipToMany"},
            },
        },
        "relationshipLinks": {
            "description": "optional references to other resource objects",
            "type": "object",
            "additionalProperties": True,
            "properties": {
                "self": {"$ref": "#/components/schemas/link"},
                "related": {"$ref": "#/components/schemas/link"},
            },
        },
        "relationshipToMany": {
            "description": "An array of objects each containing the "
            "'type' and 'id' for to-many relationships",
            "type": "array",
            "items": {"$ref": "#/components/schemas/linkage"},
            "uniqueItems": True,
        },
        # A RelationshipView uses a ResourceIdentifierObjectSerializer
        # (hence the name ResourceIdentifierObject returned
        # by get_component_name()) which serializes type and
        # id. These can be lists or individual items depending on whether the
        # relationship is toMany or toOne so offer both options since we are
        # not iterating over all the possible {related_field}'s but rather
        # rendering one path schema which may represent to Many and toOne
        # relationships.
        "ResourceIdentifierObject": {
            "oneOf": [
                {"$ref": "#/components/schemas/relationshipToOne"},
                {"$ref": "#/components/schemas/relationshipToMany"},
            ]
        },
        "linkage": {
            "type": "object",
            "description": "the 'type' and 'id'",
            "required": ["type", "id"],
            "properties": {
                "type": {"$ref": "#/components/schemas/type"},
                "id": {"$ref": "#/components/schemas/id"},
                "meta": {"$ref": "#/components/schemas/meta"},
            },
        },
        "pagination": {
            "type": "object",
            "properties": {
                "first": {"$ref": "#/components/schemas/pageref"},
                "last": {"$ref": "#/components/schemas/pageref"},
                "prev": {"$ref": "#/components/schemas/pageref"},
                "next": {"$ref": "#/components/schemas/pageref"},
            },
        },
        "pageref": {
            "oneOf": [
                {"type": "string", "format": "uri-reference"},
                {"$ref": "#/components/schemas/nulltype"},
            ]
        },
        "failure": {
            "type": "object",
            "required": ["errors"],
            "properties": {
                "errors": {"$ref": "#/components/schemas/errors"},
                "meta": {"$ref": "#/components/schemas/meta"},
                "jsonapi": {"$ref": "#/components/schemas/jsonapi"},
                "links": {"$ref": "#/components/schemas/links"},
            },
        },
        "errors": {
            "type": "array",
            "items": {"$ref": "#/components/schemas/error"},
            "uniqueItems": True,
        },
        "error": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "id": {"type": "string"},
                "status": {"type": "string"},
                "links": {"$ref": "#/components/schemas/links"},
                "code": {"type": "string"},
                "title": {"type": "string"},
                "detail": {"type": "string"},
                "source": {
                    "type": "object",
                    "properties": {
                        "pointer": {
                            "type": "string",
                            "description": "A [JSON Pointer](https://tools.ietf.org/html/rfc6901) "  # noqa
                            "to the associated entity in the request document "
                            "[e.g. `/data` for a primary data object, or "
                            "`/data/attributes/title` for a specific attribute.",  # noqa
                        },
                        "parameter": {
                            "type": "string",
                            "description": "A string indicating which query parameter "  # noqa
                            "caused the error.",
                        },
                        "meta": {"$ref": "#/components/schemas/meta"},
                    },
                },
            },
        },
        "onlymeta": {
            "additionalProperties": False,
            "properties": {"meta": {"$ref": "#/components/schemas/meta"}},
        },
        "meta": {"type": "object", "additionalProperties": True},
        "datum": {
            "description": "singular item",
            "properties": {"data": {"$ref": "#/components/schemas/resource"}},
        },
        "nulltype": {"type": "object", "nullable": True, "default": None},
        "type": {
            "type": "string",
            "description": "The [type]"
            "(https://jsonapi.org/format/#document-resource-object-identification) "  # noqa
            "member is used to describe resource objects that share common attributes "  # noqa
            "and relationships.",
        },
        "id": {
            "type": "string",
            "description": "Each resource object’s type and id pair MUST "
            "[identify]"
            "(https://jsonapi.org/format/#document-resource-object-identification) "  # noqa
            "a single, unique resource.",
        },
    },
    "parameters": {
        "include": {
            "name": "include",
            "in": "query",
            "description": "[list of included related resources]"
            "(https://jsonapi.org/format/#fetching-includes)",
            "required": False,
            "style": "form",
            "schema": {"type": "string"},
        },
        # TODO: deepObject not well defined/supported:
        #       https://github.com/OAI/OpenAPI-Specification/issues/1706
        "fields": {
            "name": "fields",
            "in": "query",
            "description": "[sparse fieldsets]"
            "(https://jsonapi.org/format/#fetching-sparse-fieldsets).\n"
            "Use fields[\\<typename\\>]=field1,field2,...,fieldN",
            "required": False,
            "style": "deepObject",
            "schema": {
                "type": "object",
            },
            "explode": True,
        },
        "sort": {
            "name": "sort",
            "in": "query",
            "description": "[list of fields to sort by]"
            "(https://jsonapi.org/format/#fetching-sorting)",
            "required": False,
            "style": "form",
            "schema": {"type": "string"},
        },
    },
}



import uuid

from django.contrib.auth.models import AbstractUser, Permission
from django.db import models

from papermerge.core.models.document import Document
from papermerge.core.models.folder import Folder
from papermerge.core.models.node import BaseTreeNode
from papermerge.core.models.page import Page
from papermerge.core.models.tags import (
    ColoredTag,
    Tag
)

from papermerge.core.models.document_version import DocumentVersion


class User(AbstractUser):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)

    # Initially user is created with empty `home_folder` and `inbox_folder`.
    # Home and Inbox folder fields are populated as part of `post_save` model
    # `user` signal
    home_folder = models.OneToOneField(
        'Folder',
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name='home_folder_of'
    )
    inbox_folder = models.OneToOneField(
        'Folder',
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name='inbox_folder_of'
    )

    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        auto_now=True
    )

    def create_special_folders(self):
        """
        Creates user's home and inbox folders.

        This method is invoked in user's post save
        signal on model creation.
        """
        _inbox, _ = Folder.objects.get_or_create(
            title=Folder.INBOX_TITLE,
            parent=None,
            user=self
        )
        _home, _ = Folder.objects.get_or_create(
            title=Folder.HOME_TITLE,
            parent=None,
            user=self
        )
        self.inbox_folder = _inbox
        self.home_folder = _home
        self.save()

    @property
    def perm_codenames(self):
        """
        Returns aggregated list of permissions codenames of the user.

        Permissions are aggregated from self.user_permissions +
        permissions from all self.groups.
        In Django permissions can be assigned directly to user object
        or via groups. This attribute returns list of string codenames
        of all permissions - from user object and from each associated
        group. It is meant to be passed (and used by) the frontend
        in order to toggle on/off different parts depending on what
        user is allowed/not allowed to view/perform.

        Note that codename DOES NOT include app label.
        """

        # 1. gather all perms via associated groups
        user_groups_field = self._meta.get_field('groups')
        user_groups_query = 'group__%s' % user_groups_field.related_query_name()
        result1 = Permission.objects.filter(
            **{user_groups_query: self}
        ).values_list(
            'codename',
            flat=True
        )
        # 2. gather all perms via user obj
        result2 = self.user_permissions.values_list(
            'codename',
            flat=True
        )

        # combine 1. and 2.
        result = list(result1) + list(result2)

        # remove all duplicates
        return list(set(result))

    def delete(self, using=None, keep_parents=False):
        Document.objects.filter(user=self).delete()
        super().delete(using=using, keep_parents=keep_parents)



import magic

from rest_framework.exceptions import PermissionDenied
from rest_framework.generics import RetrieveAPIView, GenericAPIView

from django.http import (
    Http404,
    HttpResponse
)

from papermerge.core.models import DocumentVersion
from papermerge.core.serializers import (
    DocumentVersionSerializer,
    DocumentVersionDownloadSerializer
)
from papermerge.core.renderers import PDFRenderer
from .mixins import RequireAuthMixin


class DocumentVersionsDownloadView(RequireAuthMixin, GenericAPIView):

    serializer_class = DocumentVersionDownloadSerializer
    renderer_classes = [PDFRenderer]

    def get(self, *args, **kwargs):
        doc_ver = self.get_object()

        file_abs_path = doc_ver.abs_file_path()

        mime_type = magic.from_file(file_abs_path, mime=True)
        try:
            file_handle = open(file_abs_path, "rb")
        except OSError:
            raise Http404("Cannot open local version of the document")

        resp = HttpResponse(
            file_handle.read(),
            content_type=mime_type
        )
        disposition = "attachment; filename=%s" % doc_ver.document.title
        resp['Content-Disposition'] = disposition
        file_handle.close()

        return resp

    def get_object(self):
        doc_ver = DocumentVersion.objects.get(
            pk=self.kwargs['pk']
        )
        if doc_ver.document.user != self.request.user:
            raise PermissionDenied

        return doc_ver


class DocumentVersionView(RequireAuthMixin, RetrieveAPIView):
    serializer_class = DocumentVersionSerializer

    def get_queryset(self):
        return DocumentVersion.objects.filter(
            document__user=self.request.user
        )



# Generated by Django 4.0.8 on 2022-12-22 06:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0004_remove_automate_dst_folder_remove_automate_tags_and_more'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='basetreenode',
            constraint=models.UniqueConstraint(fields=('parent', 'title'), name='unique title per parent'),
        ),
    ]



from .document import DocumentSerializer
from .document import (
    DocumentDetailsSerializer,
    DocumentsMergeSerializer
)
from .document_version import (
    DocumentVersionSerializer,
    DocumentVersionOcrTextSerializer,
    DocumentVersionDownloadSerializer
)
from .folder import FolderSerializer
from .node import (
    NodeSerializer,
    NodeMoveSerializer,
    NodesDownloadSerializer,
    NodeTagsSerializer,
    InboxCountSerializer
)
from .ocr import OcrSerializer
from .user import UserSerializer, Data_UserSerializer
from .group import GroupSerializer
from .tag import TagSerializer
from .password import PasswordSerializer
from .permission import PermissionSerializer
from .page import (
    PageSerializer,
    PageDeleteSerializer,
    PagesReorderSerializer,
    PageReorderSerializer,
    PagesRotateSerializer,
    PageRotateSerializer,
    PagesMoveToFolderSerializer,
    PagesMoveToDocumentSerializer
)
from .content_type import ContentTypeSerializer
from .preferences import CustomUserPreferenceSerializer
from .token import (
    TokenSerializer,
    CreateTokenSerializer
)
from .version import VersionSerializer



from functools import wraps

from django.core.exceptions import PermissionDenied


def permission_required(name):
    """
    Grants access to DRF View method only if user has permission

    Example of usage:

        class SomeView(GenericAPIView):

            @permission_required('change_group')
            def post(self, request):
                # reaches this place only if user has `change_group` perm
                pass
    """
    def wrap(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            if hasattr(self, 'request'):
                if hasattr(self.request, 'user'):
                    if self.request.user.has_perm(f"core.{name}"):
                        return func(self, *args, **kwargs)

            raise PermissionDenied

        return inner

    return wrap



import pytz
import uuid

from django.utils import timezone
from django.db import models

from taggit.managers import TaggableManager
from taggit.managers import _TaggableManager

from papermerge.core import validators
from papermerge.core.models.tags import ColoredTag

from .utils import uuid2raw_str

NODE_TYPE_FOLDER = 'folder'
NODE_TYPE_DOCUMENT = 'document'


def move_node(source_node, target_node):
    source_node.parent = target_node
    source_node.save()


class PolymorphicTagManager(_TaggableManager):
    """
    What is this ugliness all about?

    `taggit` adds tags to models. Besides useful attributes of tag like
    name and color, tags also consider the "type" of the associated model.
    For example if we would add tag to Folder model - f1.add(['red', 'blue'])
    when looking up tags associated to f1 i.e. `f1.tags.all()` `taggit`
    internals will search for all tags with name 'red' and 'blue' AND
    model name `Folder` (actually django's ContentType of the model `Folder`).
    Similar for `Document` model doc.tags.all() - will look up for all tags
    associated to doc instance AND model `Document`.

    In context of Papermerge, both `Folder` and `Documents` are `BaseTreeNode`s
    as well - so when user adds tags to `Folder` or `Document` instances he/she
    expects to find same tags when looking via associated node instances.

    Example A:

        $ f1 = Folder.objects.create(...)
        $ f1.tags.add(['red', 'blue'])
        $ node = BaseTreeNode.objects.get(pk=f1.pk)     (1)
        $ node.tags.all() == f1.tags.all()              (2)

    In (1) we get associated node of the folder f1 - and in (2) we expect
    that node instance will have tags 'red' and 'blue' associated.

    The problem is that `taggit` does not work that way and without
    workaround implemented by `PolymorphicTagManager` the scenario described
    in Example A will not work - because when performing `node.tags.all()`
    default `taggit` behaviour is to consider `ContentType` of the associated
    instance - in this case `BaseTreeNode`; because tags were added via Folder
    - they won't be found when looked up via `BaseTreeNode` (and vice versa).
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Instead of Document or Folder instances/models
        # always use associated BaseTreeNode instances/model
        if hasattr(self.instance, 'basetreenode_ptr'):  # Document or Folder
            self.model = self.instance.basetreenode_ptr.__class__
            self.instance = self.instance.basetreenode_ptr


class NodeManager(models.Manager):
    pass


class NodeQuerySet(models.QuerySet):

    def delete(self, *args, **kwargs):
        for node in self:
            descendants = node.get_descendants()

            if descendants.count() > 0:
                descendants.delete(*args, **kwargs)
            # At this point all descendants were deleted.
            # Self delete :)
            try:
                node.delete(*args, **kwargs)
            except BaseTreeNode.DoesNotExist:
                # this node was deleted by earlier recursive call
                # it is ok, just skip
                pass


CustomNodeManager = NodeManager.from_queryset(NodeQuerySet)


class BaseTreeNode(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)

    parent = models.ForeignKey(
        'self',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name='children',
        verbose_name='parent'
    )
    # shortcut - helps to figure out if node is either folder or document
    # without performing extra joins. This field may be empty. In such
    # case you need to perform joins with Folder/Document table to figure
    # out what type of node is this instance.
    ctype = models.CharField(
        max_length=16,
        choices=(
            (NODE_TYPE_FOLDER, NODE_TYPE_FOLDER),
            (NODE_TYPE_DOCUMENT, NODE_TYPE_DOCUMENT),
        ),
        blank=True,
        null=True
    )

    title = models.CharField(
        "Title",
        max_length=200,
        validators=[validators.safe_character_validator]
    )

    lang = models.CharField(
        'Language',
        max_length=8,
        blank=False,
        null=False,
        default='deu'
    )

    user = models.ForeignKey(
        'User',
        on_delete=models.CASCADE
    )

    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        auto_now=True
    )

    tags = TaggableManager(
        through=ColoredTag,
        manager=PolymorphicTagManager
    )

    # custom Manager + custom QuerySet
    objects = CustomNodeManager()

    @property
    def idified_title(self):
        """
        Returns a title with ID part inserted at the end

        Example:
            input: title="My Invoices", id="233453"
            output: "My Invoices-233453"
        """

        return f'{self.title}-{self.id}'

    def human_datetime(self, _datetime) -> str:
        """
        Localize and format datetime instance considering user preferences.
        """
        tz = pytz.timezone(
            self.user.preferences['localization__timezone']
        )
        fmt = self.user.preferences['localization__date_format']
        fmt += " " + self.user.preferences['localization__time_format']

        ret_datetime = timezone.localtime(_datetime, timezone=tz)

        ret = ret_datetime.strftime(fmt)
        return ret

    @property
    def human_updated_at(self) -> str:
        """
        updated_at displayed considering user's timezone, date and time prefs.

        returns string with user friendly formated datetime.
        """

        ret = self.human_datetime(self.updated_at)
        return ret

    @property
    def human_created_at(self) -> str:
        """
        created_at displayed considering user's timezone, date and time prefs.
        """
        ret = self.human_datetime(self.created_at)
        return ret

    @property
    def _type(self) -> str:
        try:
            self.folder
        except Exception:
            return NODE_TYPE_DOCUMENT

        return NODE_TYPE_FOLDER

    @property
    def document_or_folder(self):
        """Returns instance of associated `Folder` or `Document` model"""
        return self.folder_or_document

    @property
    def folder_or_document(self):
        """Returns instance of associated `Folder` or `Document` model"""
        if self.is_folder:
            return self.folder

        return self.document

    @property
    def is_folder(self) -> bool:
        """Returns True if and only if this node is a folder"""
        if self.ctype in (NODE_TYPE_DOCUMENT, NODE_TYPE_FOLDER):
            return self.ctype == NODE_TYPE_FOLDER

        return self._type == NODE_TYPE_FOLDER

    @property
    def is_document(self) -> bool:
        """Returns True if and only if this node is a document"""
        if self.ctype in (NODE_TYPE_DOCUMENT, NODE_TYPE_FOLDER):
            return self.ctype == NODE_TYPE_DOCUMENT

        return self._type == NODE_TYPE_DOCUMENT

    def get_ancestors(self, include_self=True):
        """Returns all ancestors of the node"""
        sql = '''
        WITH RECURSIVE tree AS (
            SELECT *, 0 as level FROM core_basetreenode WHERE id = %s
            UNION ALL
            SELECT core_basetreenode.*, level + 1
            FROM core_basetreenode, tree
            WHERE core_basetreenode.id = tree.parent_id
        )
        '''
        node_id = uuid2raw_str(self.pk)
        if include_self:
            sql += 'SELECT * FROM tree ORDER BY level DESC'
            return BaseTreeNode.objects.raw(sql, [node_id])

        sql += 'SELECT * FROM tree WHERE NOT id = %s ORDER BY level DESC'

        return BaseTreeNode.objects.raw(sql, [node_id, node_id])

    def get_descendants(self, include_self=True):
        """Returns all descendants of the node"""
        sql = '''
        WITH RECURSIVE tree AS (
            SELECT * FROM core_basetreenode WHERE id = %s
            UNION ALL
            SELECT core_basetreenode.* FROM core_basetreenode, tree
              WHERE core_basetreenode.parent_id = tree.id
        )
        '''
        node_id = uuid2raw_str(self.pk)
        if include_self:
            sql += 'SELECT * FROM tree'
            return BaseTreeNode.objects.raw(sql, [node_id])

        sql += 'SELECT * FROM tree WHERE NOT id = %s'
        return BaseTreeNode.objects.raw(sql, [node_id, node_id])

    def save(self, *args, **kwargs):
        if not self.ctype:
            self.ctype = self.__class__.__name__.lower()
        return super().save(*args, **kwargs)

    class Meta:
        # please do not confuse this "Documents" verbose name
        # with real Document object, which is derived from BaseNodeTree.
        # The reason for this naming confusing is that from the point
        # of view of users, the BaseNodeTree are just a list of documents.
        verbose_name = "Documents"
        verbose_name_plural = "Documents"
        _icon_name = 'basetreenode'

        constraints = [
            models.UniqueConstraint(
                fields=['parent', 'title'], name='unique title per parent'
            ),
        ]
